""" MultiQC submodule to parse output from Picard ExtractIlluminaBarcodes """

import logging
from collections import defaultdict

from multiqc.modules.picard import util
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard ExtractIlluminaBarcodes reports and parse their data"""

    data_by_lane = defaultdict(dict)

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/extractilluminabarcodes", filehandles=True):
        # Sample name from input file name by default
        lane = f["s_name"]
        keys = None

        for line in f["f"]:
            maybe_lane_name = util.extract_sample_name(
                module, line, f, picard_tool="ExtractIlluminaBarcodes", picard_opt="LANE"
            )
            if maybe_lane_name:
                # Starts information for a new sample
                lane = maybe_lane_name
                keys = None

            if util.is_line_right_before_table(line, picard_class=["ExtractIlluminaBarcodes", "BarcodeMetric"]):
                keys = f["f"].readline().strip("\n").split("\t")
                module.add_data_source(f, s_name=lane, section="ExtractIlluminaBarcodes")

            elif keys:
                vals = line.strip("\n").split("\t")
                if len(vals) != len(keys):
                    keys = None
                    continue

                data = dict(zip(keys, vals))
                data["LANE"] = lane
                data_by_lane[lane][data["BARCODE"]] = data

    data_by_lane = module.ignore_samples(data_by_lane)
    if len(data_by_lane) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_lane, f"multiqc_{module.anchor}_ExtractIlluminaBarcodes")

    plot_data = {"per_lane": reads_per_lane(data_by_lane), "per_bc": reads_per_barcode(data_by_lane)}

    per_lane_plot_config = {
        "id": f"plot-{module.anchor}-illuminabarcodemetrics-readsperlane",
        "title": f"{module.name} ExtractIlluminaBarcodes: Reads per lane",
        "ylab": "Lane",
        "data_labels": [
            {"name": "Reads", "ylab": "Number of Reads"},
            {"name": "Matches", "ylab": "Number of Matching Reads"},
            {"name": "PF Matches", "ylab": "Number of Passing Filter Reads"},
        ],
    }
    per_barcode_plot_config = {
        "id": f"plot-{module.anchor}-illuminabarcodemetrics-readsperbarcode",
        "title": f"{module.name} ExtractIlluminaBarcodes: Reads per barcode",
        "ylab": "Lane",
        "data_labels": [
            {"name": "Reads", "ylab": "Number of Reads"},
            {"name": "Matches", "ylab": "Number of Matching Reads"},
            {"name": "PF Matches", "ylab": "Number of Passing Filter Reads"},
        ],
    }

    plot_cats = [dict(), dict(), dict()]
    plot_cats[0]["READS"] = {"name": "Reads"}
    plot_cats[1]["PERFECT_MATCHES"] = {"name": "Perfect Matching Reads"}
    plot_cats[1]["ONE_MISMATCH_MATCHES"] = {"name": "One Mismatch Reads"}
    plot_cats[2]["PF_PERFECT_MATCHES"] = {"name": "Perfect Matching Passing Filter Reads"}
    plot_cats[2]["PF_ONE_MISMATCH_MATCHES"] = {"name": "Passing Filter One Mismatch Reads"}

    module.add_section(
        name="Barcode Metrics Per Lane",
        anchor=f"{module.anchor}-illuminabarcodemetrics-perlane",
        description="""
        Indicates the number of matches (and mismatches) between the barcode reads and the actual barcodes.
        See the [Picard Documentation](https://broadinstitute.github.io/picard/picard-metric-definitions.html#ExtractIlluminaBarcodes.BarcodeMetric) for details.
        """,
        plot=bargraph.plot(
            [plot_data["per_lane"], plot_data["per_lane"], plot_data["per_lane"]],
            plot_cats,
            per_lane_plot_config,
        ),
    )

    module.add_section(
        name="Barcode Metrics Per Barcode",
        anchor=f"{module.anchor}-illuminabarcodemetrics-perbarcode",
        description="""
        Indicates the number of matches (and mismatches) between the barcode reads and the actual barcodes.
        See the [Picard Documentation](https://broadinstitute.github.io/picard/picard-metric-definitions.html#ExtractIlluminaBarcodes.BarcodeMetric) for details.
        """,
        plot=bargraph.plot(
            [plot_data["per_bc"], plot_data["per_bc"], plot_data["per_bc"]],
            plot_cats,
            per_barcode_plot_config,
        ),
    )
    # Return the number of detected samples to the parent module
    return len(data_by_lane)


def reads_per_barcode(data):
    """
    Return reads/barcode for entire run
    """
    reads_per_barcode = {}
    for lane, barcodes in data.items():
        for barcode, barcode_data in barcodes.items():
            if barcode not in reads_per_barcode:
                reads_per_barcode[barcode] = {
                    "READS": 0,
                    "PERFECT_MATCHES": 0,
                    "ONE_MISMATCH_MATCHES": 0,
                    "PF_READS": 0,
                    "PF_PERFECT_MATCHES": 0,
                    "PF_ONE_MISMATCH_MATCHES": 0,
                }
            reads_per_barcode[barcode]["READS"] += int(barcode_data["READS"])
            reads_per_barcode[barcode]["PERFECT_MATCHES"] += int(barcode_data["PERFECT_MATCHES"])
            reads_per_barcode[barcode]["ONE_MISMATCH_MATCHES"] += int(barcode_data["ONE_MISMATCH_MATCHES"])
            reads_per_barcode[barcode]["PF_READS"] += int(barcode_data["PF_READS"])
            reads_per_barcode[barcode]["PF_PERFECT_MATCHES"] += int(barcode_data["PF_PERFECT_MATCHES"])
            reads_per_barcode[barcode]["PF_ONE_MISMATCH_MATCHES"] += int(barcode_data["PF_ONE_MISMATCH_MATCHES"])
    return reads_per_barcode


def reads_per_lane(data):
    """
    Return total reads/lane
    """
    reads_per_lane = {}
    for lane, barcodes in data.items():
        reads_per_lane[lane] = {
            "READS": 0,
            "PERFECT_MATCHES": 0,
            "ONE_MISMATCH_MATCHES": 0,
            "PF_READS": 0,
            "PF_PERFECT_MATCHES": 0,
            "PF_ONE_MISMATCH_MATCHES": 0,
        }
        for barcode, barcode_data in barcodes.items():
            reads_per_lane[lane]["READS"] += int(barcode_data["READS"])
            reads_per_lane[lane]["PERFECT_MATCHES"] += int(barcode_data["PERFECT_MATCHES"])
            reads_per_lane[lane]["ONE_MISMATCH_MATCHES"] += int(barcode_data["ONE_MISMATCH_MATCHES"])
            reads_per_lane[lane]["PF_READS"] += int(barcode_data["PF_READS"])
            reads_per_lane[lane]["PF_PERFECT_MATCHES"] += int(barcode_data["PF_PERFECT_MATCHES"])
            reads_per_lane[lane]["PF_ONE_MISMATCH_MATCHES"] += int(barcode_data["PF_ONE_MISMATCH_MATCHES"])

    return reads_per_lane



""" MultiQC module to parse output from Space Ranger count """

import json
import logging
import os
from collections import OrderedDict

from multiqc import config
from multiqc.plots import linegraph, table

from ._utils import set_hidden_cols, transform_data, update_dict

# Initialise the logger
log = logging.getLogger(__name__)


class SpaceRangerCountMixin:
    """Space Ranger count report parser"""

    def parse_count_html(self):
        self.spacerangercount_data = dict()
        self.spacerangercount_general_data = dict()
        self.spacerangercount_warnings = dict()
        self.spacerangercount_plots_conf = {"saturation": dict(), "genes": dict(), "genomic_dna": dict()}
        self.spacerangercount_plots_data = {"saturation": dict(), "genes": dict(), "genomic_dna": dict()}
        self.count_general_data_headers = dict()
        self.count_data_headers = OrderedDict()
        self.count_warnings_headers = OrderedDict()

        for f in self.find_log_files("spaceranger/count_html", filehandles=True):
            self.parse_count_report(f)

        self.spacerangercount_data = self.ignore_samples(self.spacerangercount_data)
        self.spacerangercount_general_data = self.ignore_samples(self.spacerangercount_general_data)
        self.spacerangercount_warnings = self.ignore_samples(self.spacerangercount_warnings)
        for k in self.spacerangercount_plots_data.keys():
            self.spacerangercount_plots_data[k] = self.ignore_samples(self.spacerangercount_plots_data[k])

        self.count_general_data_headers["reads"] = {
            "rid": "count_genstats_reads",
            "title": "Reads",
            "description": f"Number of reads ({config.read_count_desc})",
            "shared_key": "read_count",
            "namespace": "Space Ranger Count",
        }

        self.count_data_headers["reads"] = {
            "rid": "count_data_reads",
            "title": "Reads",
            "description": f"Number of reads ({config.read_count_desc})",
            "shared_key": "read_count",
        }
        self.count_data_headers = set_hidden_cols(
            self.count_data_headers,
            [
                "Q30 bc",
                "Q30 UMI",
                "Q30 read",
                "reads in spots",
                "avg reads/spot",
                "confident reads",
                "confident filtered reads",
                "genomic unis/unspliced probe",
                "valid umi",
                "saturation",
            ],
        )

        if len(self.spacerangercount_general_data) == 0:
            return 0

        else:
            self.general_stats_addcols(self.spacerangercount_general_data, self.count_general_data_headers)

            # Write parsed report data to a file
            self.write_data_file(self.spacerangercount_data, "multiqc_spaceranger_count")

            # Add sections to the report
            if len(self.spacerangercount_warnings) > 0:
                self.add_section(
                    name="Count - Warnings",
                    anchor="spaceranger-count-warnings-section",
                    description="Warnings encountered during the analysis",
                    plot=table.plot(
                        self.spacerangercount_warnings,
                        self.count_warnings_headers,
                        {
                            "namespace": "Space Ranger Count",
                            "id": "spaceranger-count-warnings",
                            "title": "Space Ranger Count: Warnings",
                        },
                    ),
                )

            self.add_section(
                name="Count - Summary stats",
                anchor="spaceranger-count-stats-section",
                description="Summary QC metrics from Space Ranger count",
                plot=table.plot(
                    self.spacerangercount_data,
                    self.count_data_headers,
                    {
                        "namespace": "Space Ranger Count",
                        "id": "spaceranger-count-stats",
                        "title": "Space Ranger Count: Summary stats",
                    },
                ),
            )

            # The gDNA plots are only contained for spaceranger workflows with probesets
            if "genomic_dna" in self.spacerangercount_plots_data:
                self.add_section(
                    name="Count - UMIs from Genomic DNA",
                    anchor="spaceranger-count-bcrank-plot-section",
                    description=self.spacerangercount_plots_conf["genomic_dna"]["description"],
                    helptext=self.spacerangercount_plots_conf["genomic_dna"]["helptext"],
                    plot=linegraph.plot(
                        self.spacerangercount_plots_data["genomic_dna"],
                        self.spacerangercount_plots_conf["genomic_dna"]["config"],
                    ),
                )

            if "genes" in self.spacerangercount_plots_data:
                self.add_section(
                    name="Count - Median genes",
                    anchor="spaceranger-count-genes-plot-section",
                    description=self.spacerangercount_plots_conf["genes"]["description"],
                    helptext=self.spacerangercount_plots_conf["genes"]["helptext"],
                    plot=linegraph.plot(
                        self.spacerangercount_plots_data["genes"], self.spacerangercount_plots_conf["genes"]["config"]
                    ),
                )

            if "saturation" in self.spacerangercount_plots_data:
                self.add_section(
                    name="Count - Saturation plot",
                    anchor="spaceranger-count-saturation-plot-section",
                    description=self.spacerangercount_plots_conf["saturation"]["description"],
                    helptext=self.spacerangercount_plots_conf["saturation"]["helptext"],
                    plot=linegraph.plot(
                        self.spacerangercount_plots_data["saturation"],
                        self.spacerangercount_plots_conf["saturation"]["config"],
                    ),
                )

            return len(self.spacerangercount_general_data)

    def parse_count_report(self, f):
        """Go through the html report of space ranger and extract the data in a dicts"""

        summary = None
        for line in f["f"]:
            line = line.strip()
            if line.startswith("const data"):
                line = line.replace("const data = ", "")
                summary = json.loads(line)
                summary = summary["summary"]
                break

        assert summary is not None, "Couldn't find JSON summary data in HTML report."

        sample_name = self.clean_s_name(summary["sample"]["id"], f)
        software = next(
            iter(x[1] for x in summary["summary_tab"]["pipeline_info_table"]["rows"] if x[0] == "Pipeline Version")
        )
        software_name, software_version = software.split("-")
        self.add_software_version(version=software_version, sample=sample_name, software_name=software_name)

        # List of data collated from different tables in cellranger reports.
        # This is a list of Tuples (metric name, value)
        data_rows = (
            [["Number of Spots Under Tissue", summary["summary_tab"]["filtered_bcs_transcriptome_union"]["metric"]]]
            + summary["summary_tab"]["cells"]["table"]["rows"]
            + summary["summary_tab"]["sequencing"]["table"]["rows"]
            + summary["summary_tab"]["mapping"]["table"]["rows"]
        )
        # This is only contained in spaceranger reports for analyses with probe sets
        try:
            data_rows.extend(summary["analysis_tab"]["gdna"]["gems"]["table"]["rows"])
        except KeyError as e:
            fname = os.path.join(f["root"], f["fn"])
            log.debug(
                f"Could not parse the expected DNA table in the spaceranger report: {fname}: '{e}' field is missing"
            )

        # Store general stats
        col_dict = {
            "Number of Spots Under Tissue": "spots under tissue",
            "Mean Reads per Spot": "avg reads/spot",
            "Fraction Reads in Spots Under Tissue": "reads in spots",
            "Number of Reads": "reads",
            "Valid Barcodes": "valid bc",
        }
        colours = {
            "spots under tissue": "RdPu",
            "avg reads/spot": "Blues",
            "reads in spots": "PiYG",
            "reads": "YlGn",
            "valid bc": "RdYlGn",
        }
        int_cols = [
            "reads",
            "spots under tissue",
            "avg reads/spot",
        ]
        data_general_stats = {}
        update_dict(
            data_general_stats, self.count_general_data_headers, data_rows, col_dict, colours, "Count", int_cols
        )

        # Store full data from space ranger count report
        col_dict = {
            "Number of Reads": "reads",
            "Number of Spots Under Tissue": "spots under tissue",
            "Mean Reads per Spot": "avg reads/spot",
            "Fraction Reads in Spots Under Tissue": "reads in spots",
            "Genes Detected": "genes detected",
            "Median Genes per Spot": "median genes/spot",
            "Median UMI Counts per Spot": "median umi/spot",
            "Valid Barcodes": "valid bc",
            "Valid UMIs": "valid umi",
            "Sequencing Saturation": "saturation",
            "Q30 Bases in Barcode": "Q30 bc",
            "Q30 Bases in UMI": "Q30 UMI",
            "Q30 Bases in RNA Read": "Q30 read",
            "Reads Mapped to Probe Set": "reads mapped",
            "Reads Mapped Confidently to Probe Set": "confident reads",
            "Reads Mapped Confidently to Filtered Probe Set": "confident filtered reads",
            "Estimated UMIs from Genomic DNA": "genomic umis",
            "Estimated UMIs from Genomic DNA per Unspliced Probe": "genomic umis/unspliced probe",
        }
        colours = {
            "reads": "YlGn",
            "spots under tissue": "RdPu",
            "avg reads/spot": "Blues",
            "genes detected": "Greens",
            "median genes/spot": "Purples",
            "reads in spots": "PuBuGn",
            "valid bc": "Spectral",
            "valid umi": "RdYlGn",
            "median umi/spot": "YlGn",
            "saturation": "PRGn",
            "genomic umis": "PuRd",
            "genomic umis/unspliced probe": "YlOrRd",
        }
        int_cols = [
            "reads",
            "spots under tissue",
            "avg reads/spot",
            "median genes/spot",
            "genes detected",
            "genomic umis/unspliced probe",
        ]
        data = {}
        update_dict(data, self.count_data_headers, data_rows, col_dict, colours, "Count", int_cols)

        # Extract warnings if any
        warnings = {}
        alarms_list = summary["alarms"].get("alarms", [])
        for alarm in alarms_list:
            # "Intron mode used" alarm added in Space Ranger 7.0 lacks id
            if "id" not in alarm:
                continue
            warnings[alarm["id"]] = "FAIL"
            self.count_warnings_headers[alarm["id"]] = {
                "title": alarm["id"].replace("_", " ").title(),
                "description": alarm["title"],
                "bgcols": {"FAIL": "#f7dddc"},
            }

        # Extract data for plots
        plots = {}
        plots_data = {}
        # `analysis_tab` may not be present in the report if there are few reads
        try:
            plots["saturation"] = {
                "config": {
                    "id": "mqc_spaceranger_count_saturation",
                    "title": f"Space Ranger count: {summary['analysis_tab']['seq_saturation_plot']['help']['title']}",
                    "xlab": summary["analysis_tab"]["seq_saturation_plot"]["plot"]["layout"]["xaxis"]["title"],
                    "ylab": summary["analysis_tab"]["seq_saturation_plot"]["plot"]["layout"]["yaxis"]["title"],
                    "yLog": False,
                    "xLog": False,
                    "ymin": 0,
                    "ymax": 1,
                },
                "description": "Sequencing saturation",
                "helptext": summary["analysis_tab"]["seq_saturation_plot"]["help"]["helpText"],
            }
            plots_data["saturation"] = {
                sample_name: transform_data(summary["analysis_tab"]["seq_saturation_plot"]["plot"]["data"][0])
            }
        except KeyError as e:
            log.debug("No saturation plot found in the spaceranger report:", e)
            if "saturation" in plots:
                del plots["saturation"]
            if "saturation" in plots_data:
                del plots_data["saturation"]

        # `analysis_tab` may not be present in the report if there are few reads
        try:
            plots["genes"] = {
                "config": {
                    "id": "mqc_spaceranger_count_genesXspot",
                    "title": f"Space Ranger count: {summary['analysis_tab']['median_gene_plot']['help']['title']}",
                    "xlab": summary["analysis_tab"]["median_gene_plot"]["plot"]["layout"]["xaxis"]["title"],
                    "ylab": summary["analysis_tab"]["median_gene_plot"]["plot"]["layout"]["yaxis"]["title"],
                    "yLog": False,
                    "xLog": False,
                },
                "description": "Median gene counts per spot",
                "helptext": summary["analysis_tab"]["median_gene_plot"]["help"]["helpText"],
            }
            plots_data["genes"] = {
                sample_name: transform_data(summary["analysis_tab"]["median_gene_plot"]["plot"]["data"][0])
            }
        except KeyError:
            log.debug("No median gene plot found in the spaceranger report")
            if "genes" in plots:
                del plots["genes"]
            if "genes" in plots_data:
                del plots_data["genes"]

        # The gDNA plots are only contained for spaceranger workflows with probesets
        try:
            plots["genomic_dna"] = {
                "config": {
                    "id": "mqc_spaceranger_count_genomic_dna",
                    "title": f"Space Ranger count: {summary['analysis_tab']['gdna']['gems']['help']['title']}",
                    "xlab": summary["analysis_tab"]["gdna"]["plot"]["layout"]["xaxis"]["title"],
                    "ylab": summary["analysis_tab"]["gdna"]["plot"]["layout"]["yaxis"]["title"],
                    "yLog": False,
                    "xLog": False,
                },
                "description": "Estimated UMIs from Genomic DNA per Unspliced Probe",
                "helptext": summary["analysis_tab"]["gdna"]["gems"]["help"]["data"][2][1][0]
                + "\n\nThis summary graphic in the MultiQC report only shows the estimated mean baseline level of unspliced probe counts.",
            }
            plots_data["genomic_dna"] = {
                sample_name: transform_data(summary["analysis_tab"]["gdna"]["plot"]["data"][2])
            }
        except KeyError:
            log.debug("No genomic DNA plot found in the spaceranger report")
            if "genomic_dna" in plots:
                del plots["genomic_dna"]
            if "genomic_dna" in plots_data:
                del plots_data["genomic_dna"]

        if len(data) > 0:
            if sample_name in self.spacerangercount_general_data:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {sample_name}")
            self.add_data_source(f, sample_name, module="spaceranger", section="count")
            self.spacerangercount_data[sample_name] = data
            self.spacerangercount_general_data[sample_name] = data_general_stats
            if len(warnings) > 0:
                self.spacerangercount_warnings[sample_name] = warnings
            self.spacerangercount_plots_conf = plots
            for k in plots_data.keys():
                if k not in self.spacerangercount_plots_data.keys():
                    self.spacerangercount_plots_data[k] = dict()
                self.spacerangercount_plots_data[k].update(plots_data[k])



""" MultiQC module to parse output from ClipAndMerge """


import logging
import os
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """ClipAndMerge module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="ClipAndMerge",
            anchor="clipandmerge",
            href="http://www.github.com/apeltzer/ClipAndMerge",
            info="is a tool for adapter clipping and read merging for ancient DNA data.",
            doi="10.1186/s13059-016-0918-z",
        )

        # Find and load any ClipAndMerge reports
        self.clipandmerge_data = dict()
        for f in self.find_log_files("clipandmerge"):
            self.parse_clipandmerge_log(f)

        # Filter to strip out ignored sample names
        self.clipandmerge_data = self.ignore_samples(self.clipandmerge_data)

        if len(self.clipandmerge_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.clipandmerge_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.clipandmerge_data, "multiqc_clipandmerge")

        # Basic Stats Table
        self.clipandmerge_general_stats_table()

        # Alignment Rate Plot
        self.add_section(plot=self.clipandmerge_alignment_plot())

    def parse_clipandmerge_log(self, f):
        regexes = {
            "usable_reads": r"Number of usable reads in the output file\(s\):\s+(\d+)",
            "merged_reads": r"Number of usable merged reads:\s+(\d+)",
            "percentage": r"Percentage of usable merged reads:\s+(\d+\.\d+)",
            "usable_not_merged_forward": r"Number of usable not merged forward reads:\s+(\d+)",
            "usable_not_merged_reverse": r"Number of usable not merged reverse reads:\s+(\d+)",
            "usable_forward_no_pairing_reverse": r"Number of usable forward reads with no pairing reverse read:\s+(\d+)",
            "usable_reverse_no_pairing_forward": r"Number of usable reverse reads with no pairing forward read:\s+(\d+)",
            "identifier": r"SampleID:\s+(\S+)",
            "version": r"ClipAndMerge \(v. ([\d\.]+)\)",
        }

        parsed_data = dict()
        for k, r in regexes.items():
            r_search = re.search(r, f["f"], re.MULTILINE)
            if r_search:
                if k == "versions":
                    parsed_data[k] = r_search.group(1)
                    continue

                try:
                    parsed_data[k] = float(r_search.group(1))
                except ValueError:
                    parsed_data[k] = r_search.group(1)

        if len(parsed_data) > 0:
            s_name = self.clean_s_name(os.path.basename(f["root"]), f)
            if "identifier" in parsed_data:
                s_name = self.clean_s_name(parsed_data["identifier"], f)
            self.clipandmerge_data[s_name] = parsed_data

            if "version" in parsed_data:
                self.add_software_version(parsed_data["version"], s_name)

            self.add_data_source(f)

    def clipandmerge_general_stats_table(self):
        """Take the parsed stats from the ClipAndMerge report and add it to the
        basic stats table at the top of the report"""

        headers = {
            "percentage": {
                "title": "% Merged",
                "description": "Percentage of reads merged",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "Greens",
                "format": "{:,.2f}",
            }
        }
        self.general_stats_addcols(self.clipandmerge_data, headers)

    def clipandmerge_alignment_plot(self):
        """Make the HighCharts HTML to plot the duplication rates"""

        # Specify the order of the different possible categories
        keys = {
            "merged_reads": {"name": "Merged Reads"},
            "usable_not_merged_forward": {"name": "Usable, not merged (forward)"},
            "usable_not_merged_reverse": {"name": "Usable, not merged (reverse)"},
            "usable_forward_no_pairing_reverse": {"name": "Usable forward-only"},
            "usable_reverse_no_pairing_forward": {"name": "Usable reverse-only"},
        }

        # Config for the plot
        config = {
            "id": "clipandmerge_rates",
            "title": "ClipAndMerge: Read merging results",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
            "hide_zero_cats": False,
        }

        return bargraph.plot(self.clipandmerge_data, keys, config)



#!/usr/bin/env python

""" MultiQC module to parse output from Freyja """

import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph
from multiqc.utils import mqc_colour

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Freyja",
            anchor="freyja",
            href="https://github.com/andersen-lab/Freyja",
            info="Recover relative lineage abundances from mixed SARS-CoV-2 samples.",
            doi="10.1038/s41586-022-05049-6",
        )

        # To store the summary data
        self.freyja_data = dict()

        # Parse the output files
        self.parse_summ_files()

        # Remove filtered samples
        self.freyja_data = self.ignore_samples(self.freyja_data)

        # Let MultiQC know this module found no data
        if len(self.freyja_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.freyja_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.write_data_file(self.freyja_data, "multiqc_freyja")

        top_lineages_dict = {}
        all_lineages = set()
        for s_name, sub_dict in self.freyja_data.items():
            top_lineage = max(sub_dict, key=sub_dict.get)
            top_lineage_value = sub_dict[top_lineage]
            top_lineages_dict[s_name] = {
                "Top_lineage_freyja": top_lineage,
                "Top_lineage_freyja_percentage": top_lineage_value,
            }
            all_lineages.add(top_lineage)
        for s_name, sub_dict in self.freyja_data.items():
            for lineage, val in sub_dict.items():
                if val not in all_lineages:
                    all_lineages.add(lineage)

        self.scale = mqc_colour.mqc_colour_scale("plot_defaults")
        self.general_stats_cols(top_lineages_dict, all_lineages)
        self.add_freyja_section(all_lineages)

    def parse_summ_files(self):
        """
        Parse the summary file.
        Freyja has multiple summary files, but we only need to parse the one from the demix command.
        More specifically, we only need the line that starts with "summarized".
        ...
        summarized	[('BQ.1*', 0.983), ('Omicron', 0.011), ('key', value)]
        ...
        """
        for f in self.find_log_files("freyja", filehandles=True):
            s_name = f["s_name"]
            # Read the statistics from file
            d = {}
            for line in f["f"]:
                try:
                    if line.startswith("summarized"):
                        summarized_line = line
                        summarized_line = summarized_line.strip().split("\t")[1]
                        d = eval(
                            summarized_line
                        )  # Make sure no input is corrupted and does not contain any malicious code
                        d = dict(d)
                except ValueError:
                    pass

            # There is no sample name in the log, so we use the root of the
            # file as sample name (since the filename is always stats.dat
            if s_name in self.freyja_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.freyja_data[s_name] = d
            self.add_data_source(f, s_name)

    def general_stats_cols(self, top_lineages_dict, all_lineages):
        """Add a single column displaying the most abundant lineage to the General Statistics table"""
        headers = {
            "Top_lineage_freyja": {
                "title": "Top lineage",
                "description": "The most abundant lineage in the sample",
                "bgcols": {x: self.scale.get_colour(i) for i, x in enumerate(all_lineages)},
            },
            "Top_lineage_freyja_percentage": {
                "title": "Top lineage %",
                "description": "The percentage of the most abundant lineage in the sample",
                "max": 100,
                "min": 0,
                "scale": "Blues",
                "modify": lambda x: x * 100,
                "suffix": "%",
            },
        }

        self.general_stats_addcols(top_lineages_dict, headers)

    def add_freyja_section(self, lineages):
        pconfig = {
            "id": "Freyja_plot",
            "title": "Freyja: Top lineages",
            "ylab": "relative abundance",
            "yCeiling": 1,
            "cpswitch": False,
            "cpswitch_c_active": False,
        }
        cats = {x: {"name": x, "color": self.scale.get_colour(i, lighten=1)} for i, x in enumerate(lineages)}

        self.add_section(
            name="Freyja Summary",
            anchor="freyja-summary",
            description="""
                Relative lineage abundances from mixed SARS-CoV-2 samples. Hover over the column headers for descriptions and click _Help_ for more in-depth documentation. 
                """,
            helptext="""
                The graph denotes a sum of all lineage abundances in a particular WHO designation , otherwise they are grouped into "Other".
                Lineages abundances are calculated as the number of reads that are assigned to a particular lineage. 
                Lineages and their corresponding abundances are summarized by constellation. 

                > **Note**: Lineage designation is based on the used WHO nomenclature, which could vary over time. 
                """,
            plot=bargraph.plot(self.freyja_data, cats, pconfig),
        )



# -*- coding: utf-8 -*-

""" MultiQC module to parse output files from miRTrace """


import json
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="miRTrace",
            anchor="mirtrace",
            href="https://github.com/friedlanderlab/mirtrace",
            info="is a quality control software for small RNA sequencing data developed by Friedländer lab (KTH, Sweden).",
            doi="10.1186/s13059-018-1588-9",
        )

        # Find and load miRTrace summary statistics table
        self.summary_data = dict()
        for f in self.find_log_files("mirtrace/summary"):
            self.parse_summary(f)

        # Find and load miRTrace read length table
        self.length_data = dict()
        for f in self.find_log_files("mirtrace/length"):
            self.parse_length(f)

        # Find and load miRTrace contamination statistics summary_table
        self.contamination_data = dict()
        for f in self.find_log_files("mirtrace/contaminationbasic"):
            self.parse_contamination(f)

        # Find and load miRTrace miRNA complexity table
        self.complexity_data = dict()
        for f in self.find_log_files("mirtrace/mirnacomplexity"):
            self.parse_complexity(f)

        # Filter to strip out ignored sample names
        self.summary_data = self.ignore_samples(self.summary_data)
        self.length_data = self.ignore_samples(self.length_data)
        self.contamination_data = self.ignore_samples(self.contamination_data)
        self.complexity_data = self.ignore_samples(self.complexity_data)

        # Warning when no files are found
        if (
            max(len(self.summary_data), len(self.length_data), len(self.contamination_data), len(self.complexity_data))
            == 0
        ):
            raise ModuleNoSamplesFound

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed data to a file
        self.write_data_file(self.summary_data, "multiqc_mirtrace_summary")
        self.write_data_file(self.length_data, "multiqc_mirtrace_length")
        self.write_data_file(self.contamination_data, "multiqc_mirtrace_contamination")
        self.write_data_file(self.complexity_data, "multiqc_mirtrace_complexity")

        # Report sections
        if len(self.summary_data) > 0:
            self.add_section(name="QC Plot", anchor="mirtrace_qc", plot=self.mirtrace_qc_plot())
            self.add_section(
                name="RNA Categories", anchor="mirtrace_rna_categories", plot=self.mirtrace_rna_categories()
            )

        if len(self.length_data) > 0:
            self.add_section(
                name="Read Length Distribution", anchor="mirtrace_length", plot=self.mirtrace_length_plot()
            )

        if len(self.contamination_data) > 0:
            self.add_section(
                name="Contamination Check",
                anchor="mirtrace_contamination_check",
                plot=self.mirtrace_contamination_check(),
            )

        if len(self.complexity_data) > 0:
            self.add_section(
                name="miRNA Complexity", anchor="mirtrace_complexity", plot=self.mirtrace_complexity_plot()
            )

    # Parse a miRTrace results.json file
    def parse_summary(self, f):
        try:
            cdict = json.loads(f["f"])
        except ValueError as e:
            raise e

        if "results" in cdict.keys():
            for record in cdict["results"]:
                s_name = self.clean_s_name(record["verbosename"], f)
                parsed_data = {}
                parsed_data["filename"] = record["filename"]
                parsed_data["reads_total"] = record["stats"]["allSeqsCount"]
                parsed_data["adapter_removed_length_ok"] = record["stats"]["statsQC"][4]
                parsed_data["adapter_not_detected"] = record["stats"]["statsQC"][3]
                parsed_data["length_shorter_than_18"] = record["stats"]["statsQC"][2]
                parsed_data["low_complexity"] = record["stats"]["statsQC"][1]
                parsed_data["low_phred"] = record["stats"]["statsQC"][0]
                parsed_data["reads_mirna"] = record["stats"]["statsRNAType"][0]
                parsed_data["reads_rrna"] = record["stats"]["statsRNAType"][1]
                parsed_data["reads_trna"] = record["stats"]["statsRNAType"][2]
                parsed_data["reads_artifact"] = record["stats"]["statsRNAType"][3]
                parsed_data["reads_unknown"] = record["stats"]["statsRNAType"][4]
                if s_name in self.summary_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.add_data_source(f, s_name)
                self.summary_data[s_name] = parsed_data
        else:
            log.debug(f"No valid data {f['fn']} in miRTrace summary")
            return None

    # Parse a miRTrace mirtrace-stats-length.tsv file
    def parse_length(self, f):
        header = []
        body = {}
        lines = f["f"].splitlines()
        for line in lines:
            s = line.split("\t")
            if len(header) == 0:
                if s[0] != "LENGTH":
                    log.debug(f"No valid data {f['fn']} for read length distribution")
                    return None
                header = s[1:]
            else:
                body[s[0]] = s[1 : len(s)]

        for record in header[0 : len(header)]:
            s_name = self.clean_s_name(record, f)
            parsed_data = {}
            idx = header[0 : len(header)].index(record)
            for length in body:
                parsed_data[length] = int(body[length][idx])
            if s_name in self.length_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, s_name)
            self.length_data[s_name] = parsed_data

    # Parse a miRTrace mirtrace-stats-contamination_basic.tsv file
    def parse_contamination(self, f):
        header = []
        body = {}
        lines = f["f"].splitlines()
        for line in lines:
            s = line.split("\t")
            if len(header) == 0:
                if s[0] != "CLADE":
                    log.debug(f"No valid data {f['fn']} for contamination check")
                    return None
                header = s[1:]
            else:
                body[s[0]] = s[1 : len(s)]

        for record in header[0 : len(header)]:
            s_name = self.clean_s_name(record, f)
            parsed_data = {}
            idx = header[0 : len(header)].index(record)
            for clade in body:
                parsed_data[clade] = int(body[clade][idx])
            if s_name in self.contamination_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, s_name)
            self.contamination_data[s_name] = parsed_data

    # Parse a miRTrace mirtrace-stats-mirna-complexity.tsv file
    def parse_complexity(self, f):
        header = []
        body = {}
        lines = f["f"].splitlines()
        for line in lines:
            s = line.split("\t")
            if len(header) == 0:
                if s[0] != "DISTINCT_MIRNA_HAIRPINS_ACCUMULATED_COUNT":
                    log.debug(f"No valid data {f['fn']} for miRNA complexity")
                    return None
                header = s[1:]
            else:
                body[s[0]] = s[1 : len(s)]

        for record in header[0 : len(header)]:
            s_name = self.clean_s_name(record, f)
            parsed_data = {}
            idx = header[0 : len(header)].index(record)
            for depth in body:
                parsed_data[depth] = int(body[depth][idx]) if body[depth][idx] else 0
            if s_name in self.complexity_data:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, s_name)
            self.complexity_data[s_name] = parsed_data

    # miRTrace QC Plot
    def mirtrace_qc_plot(self):
        """Generate the miRTrace QC Plot"""

        # Specify the order of the different possible categories
        keys = {
            "adapter_removed_length_ok": {"color": "#006837", "name": "Reads ≥ 18 nt after adapter removal"},
            "adapter_not_detected": {"color": "#66bd63", "name": "Reads without adapter"},
            "length_shorter_than_18": {"color": "#fdae61", "name": "Reads < 18 nt after adapter removal"},
            "low_complexity": {"color": "#d73027", "name": "Reads with low complexity"},
            "low_phred": {"color": "#a50026", "name": "Reads with low PHRED score"},
        }

        # Config for the plot
        config = {
            "id": "mirtrace_qc_plot",
            "title": "miRTrace: QC Plot",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        return bargraph.plot(self.summary_data, keys, config)

    # miRTrace Read Length Distribution
    def mirtrace_length_plot(self):
        """Generate the miRTrace Read Length Distribution"""

        data = dict()
        for s_name in self.length_data:
            try:
                data[s_name] = {int(d): int(self.length_data[s_name][d]) for d in self.length_data[s_name]}
            except KeyError:
                pass
        if len(data) == 0:
            log.debug("No valid data for read length distribution")
            return None

        config = {
            "id": "mirtrace_length_plot",
            "title": "miRTrace: Read Length Distribution",
            "ylab": "Read Count",
            "xlab": "Read Lenth (bp)",
            "ymin": 0,
            "xmin": 0,
            "xDecimals": False,
            "tt_label": "<b>Read Length (bp) {point.x}</b>: {point.y} Read Count",
            "xPlotBands": [
                {"from": 40, "to": 50, "color": "#ffebd1"},
                {"from": 26, "to": 40, "color": "#e2f5ff"},
                {"from": 18, "to": 26, "color": "#e5fce0"},
                {"from": 0, "to": 18, "color": "#ffffe2"},
            ],
        }

        return linegraph.plot(data, config)

    # miRTrace RNA Categories
    def mirtrace_rna_categories(self):
        """Generate the miRTrace RNA Categories"""

        # Specify the order of the different possible categories
        keys = {
            "reads_mirna": {"color": "#33a02c", "name": "miRNA"},
            "reads_rrna": {"color": "#ff7f00", "name": "rRNA"},
            "reads_trna": {"color": "#1f78b4", "name": "tRNA"},
            "reads_artifact": {"color": "#fb9a99", "name": "Artifact"},
            "reads_unknown": {"color": "#d9d9d9", "name": "Unknown"},
        }

        # Config for the plot
        config = {
            "id": "mirtrace_rna_categories_plot",
            "title": "miRTrace: RNA Categories",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        return bargraph.plot(self.summary_data, keys, config)

    # miRTrace Contamination Check
    def mirtrace_contamination_check(self):
        """Generate the miRTrace Contamination Check"""

        # A library of 24 colors. Should be enough for this plot
        color_lib = [
            "#A6CEE3",
            "#1F78B4",
            "#B2DF8A",
            "#33A02C",
            "#FB9A99",
            "#E31A1C",
            "#FDBF6F",
            "#FF7F00",
            "#CAB2D6",
            "#6A3D9A",
            "#FFFF99",
            "#B15928",
            "#8DD3C7",
            "#FFFFB3",
            "#BEBADA",
            "#FB8072",
            "#80B1D3",
            "#FDB462",
            "#B3DE69",
            "#FCCDE5",
            "#D9D9D9",
            "#BC80BD",
            "#CCEBC5",
            "#FFED6F",
        ]

        idx = 0

        # Specify the order of the different possible categories
        keys = {}
        for clade in self.contamination_data[list(self.contamination_data.keys())[0]]:
            keys[clade] = {"color": color_lib[idx], "name": clade}
            if idx < 23:
                idx += 1
            else:
                idx = 0

        # Config for the plot
        config = {
            "cpswitch_c_active": False,
            "id": "mirtrace_contamination_check_plot",
            "title": "miRTrace: Contamination Check",
            "ylab": "# miRNA detected",
            "cpswitch_counts_label": "Number of detected miRNA",
        }

        return bargraph.plot(self.contamination_data, keys, config)

    # miRTrace Read Length Distribution
    def mirtrace_complexity_plot(self):
        """Generate the miRTrace miRNA Complexity Plot"""

        data = dict()
        for s_name in self.complexity_data:
            try:
                data[s_name] = {int(self.complexity_data[s_name][d]): int(d) for d in self.complexity_data[s_name]}
            except KeyError:
                pass
        if len(data) == 0:
            log.debug("No valid data for miRNA complexity")
            return None

        config = {
            "id": "mirtrace_complexity_plot",
            "title": "miRTrace: miRNA Complexity Plot",
            "ylab": "Distinct miRNA Count",
            "xlab": "Number of Sequencing Reads",
            "ymin": 0,
            "xmin": 1,
            "xDecimals": False,
            "tt_label": "<b>Number of Sequencing Reads {point.x}</b>: {point.y} Distinct miRNA Count",
        }

        return linegraph.plot(data, config)



""" MultiQC module to parse output from RNA-SeQC """


import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, beeswarm, heatmap, linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="RNA-SeQC",
            anchor="rna_seqc",
            href="https://github.com/getzlab/rnaseqc",
            info="Fast, efficient RNA-Seq metrics for quality control and process optimization",
            doi=["10.1093/bioinformatics/btab135", "10.1093/bioinformatics/bts196"],
        )

        # Parse metrics from RNA-SeQC v1
        self.rna_seqc_metrics = dict()
        for f in self.find_log_files("rna_seqc/metrics_v1", filehandles=True):
            self.parse_metrics_rnaseqc_v1(f)
            self.add_data_source(f, section="v1")

        # Parse metrics from RNA-SeQC v2
        for f in self.find_log_files("rna_seqc/metrics_v2", filehandles=True):
            self.parse_metrics_rnaseqc_v2(f)
            self.add_data_source(f, section="v2")

        # Strip out ignored sample names from metrics
        self.rna_seqc_metrics = self.ignore_samples(self.rna_seqc_metrics)

        # Parse normalised coverage information.
        self.rna_seqc_norm_high_cov = dict()
        self.rna_seqc_norm_medium_cov = dict()
        self.rna_seqc_norm_low_cov = dict()
        for f in self.find_log_files("rna_seqc/coverage"):
            self.parse_coverage(f)
        self.rna_seqc_norm_high_cov = self.ignore_samples(self.rna_seqc_norm_high_cov)
        self.rna_seqc_norm_medium_cov = self.ignore_samples(self.rna_seqc_norm_medium_cov)
        self.rna_seqc_norm_low_cov = self.ignore_samples(self.rna_seqc_norm_low_cov)

        # Parse correlation matrices
        self.rna_seqc_pearson = None
        self.rna_seqc_spearman = None
        num_correlation_samples = 0
        for f in self.find_log_files("rna_seqc/correlation"):
            num_correlation_samples += self.parse_correlation(f)

        # Work out the maximum number of samples that we have across all reports
        num_found = max(
            len(self.rna_seqc_metrics),
            len(self.rna_seqc_norm_high_cov),
            len(self.rna_seqc_norm_medium_cov),
            len(self.rna_seqc_norm_low_cov),
            num_correlation_samples,
        )

        # Stop here if we didn't find anything
        if num_found == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {num_found} samples")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write metrics to a file
        self.write_data_file(self.rna_seqc_metrics, "multiqc_rna_seqc")

        # Generate sections for the report
        self.rnaseqc_general_stats()
        self.transcript_associated_plot()
        self.plot_correlation_heatmap()
        self.strand_barplot()
        self.coverage_lineplot()
        self.bam_statplot()

    def parse_metrics_rnaseqc_v1(self, f):
        """
        Parse the metrics.tsv file from RNA-SeQC version 1.x
        """
        headers = None
        for line in f["f"]:
            s = line.strip().split("\t")
            if headers is None:
                headers = s
            else:
                s_name = s[headers.index("Sample")]
                data = dict()
                for idx, h in enumerate(headers):
                    try:
                        data[h] = float(s[idx])
                    except ValueError:
                        data[h] = s[idx]
                if s_name in self.rna_seqc_metrics:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.rna_seqc_metrics[s_name] = data

    def parse_metrics_rnaseqc_v2(self, f):
        """
        Parse the *.metrics.tsv file from RNA-SeQC version 2.x
        """
        data = {}
        s_name = f["s_name"]
        for line in f["f"]:
            s = line.split("\t")
            if s[0] == "Sample":
                s_name = self.clean_s_name(s[1], f)
            if s[0] == "Total Reads":
                s[0] = "Total Read Number"
            if s[0] == "rRNA Rate":
                s[0] = "rRNA rate"
            try:
                data[s[0]] = float(s[1])
            except ValueError:
                data[s[0]] = s[1].strip()

        if s_name in self.rna_seqc_metrics:
            log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
        self.rna_seqc_metrics[s_name] = data

    def parse_coverage(self, f):
        """Parse the RNA-SeQC Normalised Coverage Files"""
        data = dict()
        s_names = None
        j = 1
        for line in f["f"].splitlines():
            s = line.strip().split("\t")
            if s_names is None:
                s_names = [self.clean_s_name(s_name, f) for s_name in s]
                for s_name in s_names:
                    data[s_name] = dict()
            else:
                for i, v in enumerate(s):
                    data[s_names[i]][j] = float(v)
                j += 1
        if f["fn"] == "meanCoverageNorm_high.txt":
            self.rna_seqc_norm_high_cov.update(data)
        elif f["fn"] == "meanCoverageNorm_medium.txt":
            self.rna_seqc_norm_medium_cov.update(data)
        elif f["fn"] == "meanCoverageNorm_low.txt":
            self.rna_seqc_norm_low_cov.update(data)

    def parse_correlation(self, f):
        """Parse RNA-SeQC correlation matrices"""
        s_names = None
        data = list()
        for line in f["f"].splitlines():
            s = line.strip().split("\t")
            if s_names is None:
                s_names = [x for x in s if x != ""]
            else:
                data.append([float(val) for val in s[1:]])

        # Filter for ignored sample names
        filtered_s_names = list()
        filtered_data = list()
        for idx, s_name in enumerate(s_names):
            s_name = self.clean_s_name(s_name, f)
            if not self.is_ignore_sample(s_name):
                filtered_s_names.append(s_name)
                filtered_data.append(data[idx])

        if f["fn"] == "corrMatrixPearson.txt":
            self.rna_seqc_pearson = (filtered_s_names, filtered_data)
        elif f["fn"] == "corrMatrixSpearman.txt":
            self.rna_seqc_spearman = (filtered_s_names, filtered_data)

        return len(filtered_s_names)

    def rnaseqc_general_stats(self):
        """
        Add key metrics to the general stats table
        """
        headers = {
            "Expression Profiling Efficiency": {
                "title": "% Expression Efficiency",
                "description": "Ratio of exonic reads to total reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
                "modify": lambda x: float(x) * 100.0,
            },
            "Genes Detected": {
                "title": "# Genes",
                "description": "Number of genes detected with at least 5 reads",
                "min": 0,
                "scale": "Blues",
                "format": "{:,.0f}",
            },
            "rRNA rate": {
                "title": "% rRNA Alignment",
                "description": "Ribosomal RNA reads per total reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "Reds",
                "modify": lambda x: float(x) * 100.0,
            },
        }
        self.general_stats_addcols(self.rna_seqc_metrics, headers)

    def transcript_associated_plot(self):
        """Plot a bargraph showing the Transcript-associated reads"""

        # Plot bar graph of groups
        keys = {
            "Exonic Rate": {"name": "Exonic", "color": "#2f7ed8"},
            "Intronic Rate": {"name": "Intronic", "color": "#8bbc21"},
            "Intergenic Rate": {"name": "Intergenic", "color": "#0d233a"},
        }

        # Config for the plot
        pconfig = {
            "id": "rna_seqc_position_plot",
            "title": "RNA-SeQC: Transcript-associated reads",
            "ylab": "Ratio of Reads",
            "cpswitch": False,
            "ymax": 1,
            "ymin": 0,
            "tt_decimals": 3,
            "cpswitch_c_active": False,
        }
        self.add_section(
            name="Transcript-associated reads",
            anchor="rna_seqc_transcript_associated",
            helptext="""
            All of the rates given are per mapped read.

            * **Exonic Rate** is the fraction mapping within exons.
            * **Intronic Rate** is the fraction mapping within introns.
            * **Intergenic Rate** is the fraction mapping in the genomic space between genes.
            """,
            plot=bargraph.plot(self.rna_seqc_metrics, keys, pconfig),
        )

    def plot_correlation_heatmap(self):
        """Return HTML for correlation heatmap"""
        data = None
        corr_type = None
        correlation_type = getattr(config, "rna_seqc", {}).get("default_correlation", "spearman")
        if self.rna_seqc_spearman is not None and correlation_type != "pearson":
            data = self.rna_seqc_spearman
            corr_type = "Spearman"
        elif self.rna_seqc_pearson is not None:
            data = self.rna_seqc_pearson
            corr_type = "Pearson"
        if data is not None:
            pconfig = {
                "id": "rna_seqc_correlation_heatmap",
                "title": f"RNA-SeQC: {corr_type} Sample Correlation",
            }
            self.add_section(
                name=f"{corr_type} Correlation",
                anchor="rseqc-rna_seqc_correlation",
                plot=heatmap.plot(data[1], data[0], data[0], pconfig),
            )

    def strand_barplot(self):
        """Plot a bargraph showing the strandedness of alignments"""
        # Plot bar graph of groups
        keys = ["End 1 Sense", "End 1 Antisense", "End 2 Sense", "End 2 Antisense"]
        # Config for the plot
        pconfig = {
            "id": "rna_seqc_strandedness_plot",
            "title": "RNA-SeQC: Strand Specificity",
            "ylab": "% Reads",
            "cpswitch_counts_label": "# Reads",
            "cpswitch_percent_label": "% Reads",
            "ymin": 0,
            "cpswitch_c_active": False,
        }
        self.add_section(
            name="Strand Specificity",
            anchor="rna_seqc_strand_specificity",
            helptext="""
            End 1/2 Sense are the number of End 1 or 2 reads that were sequenced in the sense direction.

            End 1/2 Antisense are the number of End 1 or 2 reads that were sequenced in the antisense direction
            """,
            plot=bargraph.plot(self.rna_seqc_metrics, keys, pconfig),
        )

    def coverage_lineplot(self):
        """Make HTML for coverage line plots"""
        # Add line graph to section
        data = list()
        data_labels = list()
        if len(self.rna_seqc_norm_high_cov) > 0:
            data.append(self.rna_seqc_norm_high_cov)
            data_labels.append({"name": "High Expressed"})
        if len(self.rna_seqc_norm_medium_cov) > 0:
            data.append(self.rna_seqc_norm_medium_cov)
            data_labels.append({"name": "Medium Expressed"})
        if len(self.rna_seqc_norm_low_cov) > 0:
            data.append(self.rna_seqc_norm_low_cov)
            data_labels.append({"name": "Low Expressed"})
        pconfig = {
            "id": "rna_seqc_gene_body_coverage_plot",
            "title": "RNA-SeQC: Gene Body Coverage",
            "ylab": "% Coverage",
            "xlab": "Gene Body Percentile (5' -> 3')",
            "xmin": 0,
            "xmax": 100,
            "tt_label": "<strong>{point.x}% from 5'</strong>: {point.y:.2f}",
            "data_labels": data_labels,
        }
        if len(data) > 0:
            self.add_section(
                name="Gene Body Coverage",
                anchor="rseqc-rna_seqc_mean_coverage",
                helptext="The metrics are calculated across the transcripts with tiered expression levels.",
                plot=linegraph.plot(data, pconfig),
            )

    def bam_statplot(self):
        pconfig = {"id": "rna_seqc_bam_stat_beeswarm", "title": "RNA-SeQC: Read metrics"}
        columns = [
            "Total Read Number",
            "Alternative Alignments",
            "Chimeric Reads",
            "Duplicate Reads",
            "End 1 Mapped Reads",
            "End 2 Mapped Reads",
            "End 1 Mismatches",
            "End 2 Mismatches",
            "End 1 Sense",
            "End 2 Sense",
            "Ambiguous Reads",
            "High Quality Reads",
            "Low Quality Reads",
            "Mapped Duplicate Reads",
            "Mapped Reads",
            "Mapped Unique Reads",
            "Non-Globin Reads",
            "Non-Globin Duplicate Reads",
            "rRNA Reads",
            "Unique Mapping, Vendor QC Passed Reads",
        ]
        keys = dict()
        for col in columns:
            keys[col] = {
                "title": col,
                "shared_key": "read_count",
                "suffix": config.read_count_prefix,
            }

        self.add_section(
            name="Read Counts",
            anchor="rna_seqc_bam_stat",
            description=f"Number of reads ({config.read_count_desc}) falling into different categories.",
            helptext="Note that many of these statistics are only available from RNA-SeQC v2.x",
            plot=beeswarm.plot(self.rna_seqc_metrics, keys, pconfig),
        )



""" MultiQC module to parse QC output from PURPLE """
# Initialise the logger
import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import table

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    PURPLE is a purity ploidy estimator. It combines B-allele frequency (BAF) from AMBER,
    read depth ratios from COBALT, somatic variants and structural variants to estimate the
    purity and copy number profile of a tumor sample.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="PURPLE",
            anchor="purple",
            href="https://github.com/hartwigmedical/hmftools/",
            info="""combines B-allele frequency (BAF), read depth ratios, somatic variants and
                    structural variant breakpoints to estimate the purity and copy number profile
                    of a tumor sample, and also predicts gender, the MSI status, tumor mutational
                    load and burden, clonality and the whole genome duplication status.""",
            doi="10.1038/s41586-019-1689-y",
        )

        data_by_sample = defaultdict(dict)

        for f in self.find_log_files("purple/qc"):
            data = _parse_purple_qc(f)
            if data is not None:
                if f["s_name"] in data_by_sample:
                    log.debug(f"Duplicate PURPLE output prefix found! Overwriting: {f['s_name']}")
                self.add_data_source(f, section="stats")
                data_by_sample[f["s_name"]].update(data)

        for f in self.find_log_files("purple/purity"):
            data = _parse_purple_purity(f)
            if data is not None:
                if f["s_name"] in data_by_sample:
                    log.debug(f"Duplicate PURPLE output prefix found! Overwriting: {f['s_name']}")
                self.add_data_source(f, section="stats")
                data_by_sample[f["s_name"]].update(data)
                self.add_software_version(data.get("version"), f["s_name"])

        # Filter to strip out ignored sample names:
        data_by_sample = self.ignore_samples(data_by_sample)
        if not data_by_sample:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(data_by_sample)} reports")

        # Write data to file
        self.write_data_file(data_by_sample, "purple")

        headers = _make_table_headers()

        self.add_section(
            name="PURPLE summary",
            anchor="purple-summary",
            description="""
            PURPLE summary. See details at the
            <a href=https://github.com/hartwigmedical/hmftools/tree/master/purity-ploidy-estimator#purity-file>
            documentation</a>.""",
            plot=table.plot(
                data_by_sample,
                headers,
                {
                    "id": "purple_summary",
                    "title": "PURPLE summary",
                },
            ),
            helptext="""
            1. **QC status**. Can fail for the following 3 reasons:
                * `FAIL_SEGMENT`: removed samples with more than 220 copy number segments unsupported
                    at either end by SV breakpoints. This step was added to remove samples with extreme GC bias,
                    with differences in depth of up to or in excess of 10x between high and low GC regions.
                    GC normalisation is unreliable when the corrections are so extreme so we filter.
                * `FAIL_GENDER`: if the AMBER and COBALT gender are inconsistent, we use the COBALT gender but
                    fail the sample.
                * `FAIL_DELETED_GENES`: we fail any sample with more than 280 deleted genes. This QC step was
                    added after observing that in a handful of samples with high MB scale positive GC bias we
                    sometimes systematically underestimate the copy number in high GC regions. This can lead us to
                    incorrectly infer homozygous loss of entire chromosomes, particularly on chromosomes 17 and 19.
            2. **Ploidy status**. Reflects how we have determined the purity of the sample:
                * `NORMAL`: could fix the purity using coverage and BAF alone
                * `HIGHLY_DIPLOID`: the fitted purity solution is highly diploid (> 95%) with a large
                    range of potential solutions, but somatic variants are unable to help either because they
                    were not supplied or because their implied purity was too low.
                * `SOMATIC`: somatic variants have improved the otherwise highly diploid solution
                * `NO_TUMOR`: PURPLE failed to find any aneuploidy and somatic variants were supplied but
                    there were fewer than 300 with observed VAF > 0.1.
            """,
        )

        gen_stat_cols = {k: v for k, v in headers.items() if k in ["ploidy", "purity"]}
        self.general_stats_addcols(data_by_sample, gen_stat_cols)


def _make_table_headers():
    headers = {
        "QCStatus": {
            "title": "QC Status",
            "description": "One of PASS, FAIL_SEGMENT, FAIL_GENDER, or FAIL_DELETED_GENES. "
            "For details, use the help button.",
            "scale": False,
        },
        "ploidy": {
            "title": "Ploidy",
            "description": "Average ploidy of the tumor sample after adjusting for purity",
            "scale": "RdYlGn",
            "min": 0,
        },
        "purity": {
            "title": "Purity",
            "description": "Purity of tumor in the sample",
            "scale": "RdYlGn",
            "min": 0,
            "max": 100,
            "suffix": "%",
            "modify": lambda x: float(x) * 100.0,
        },
        "gender": {"title": "Gender", "description": "One of MALE, FEMALE or MALE_KLINEFELTER", "scale": False},
        "status": {
            "title": "Ploidy status",
            "description": "One of NORMAL, HIGHLY_DIPLOID, SOMATIC, or NO_TUMOR. For details, use the help button.",
            "scale": False,
        },
        "polyclonalProportion": {
            "title": "Polyclonal",
            "description": "Polyclonal proportion. Proportion of copy number regions that are more than 0.25 "
            "from a whole copy number",
            "scale": "RdYlGn",
            "min": 0,
            "max": 100,
            "suffix": "%",
            "modify": lambda x: float(x) * 100.0,
        },
        "wholeGenomeDuplication": {
            "title": "WGD",
            "description": "Whole genome duplication. True if more than 10 autosomes have major allele ploidy > 1.5",
            "scale": False,
        },
        "msIndelsPerMb": {
            "title": "MS indel per Mb",
            "description": "Microsatellite indels per mega base",
            "scale": "RdYlGn",
            "hidden": True,
        },
        "msStatus": {
            "title": "MS status",
            "description": "Microsatellite status. One of MSI, MSS or UNKNOWN if somatic variants not supplied",
            "scale": False,
        },
        "tml": {
            "title": "TML",
            "description": "Tumor mutational load (# of missense variants in sample)",
            "scale": "RdYlGn",
            "hidden": True,
        },
        "tmlStatus": {
            "title": "TML status",
            "description": "Tumor mutational load status (# of missense variants in sample). One of HIGH, LOW or UNKNOWN "
            "if somatic variants not supplied",
            "scale": False,
        },
        "tmbPerMb": {
            "title": "TMB per Mb",
            "description": "Tumor mutational burden (# of passing variants) per mega base",
            "scale": "RdYlGn",
            "hidden": True,
        },
        "tmbStatus": {
            "title": "TMB status",
            "description": "Tumor mutational burden (# of passing variants per Mb) status. One of HIGH, LOW or UNKNOWN "
            "if somatic variants not supplied",
            "scale": False,
        },
    }
    return headers


def _parse_purple_qc(f):
    """
    $ cat <sample>.purple.qc
    QCStatus        FAIL_DELETED_GENES
    SegmentPass     true
    GenderPass      true
    DeletedGenesPass        false
    SegmentScore    0
    UnsupportedSegments     0
    Ploidy  2.0036
    AmberGender     MALE
    CobaltGender    MALE
    DeletedGenes    5529
    """

    data = dict()
    for line in f["f"].splitlines():
        fields = line.strip().split("\t")
        if len(fields) == 2:
            data[fields[0]] = fields[1]
    return data


def _parse_purple_purity(f):
    """
    $ cat <sample>.purple.purity.tsv
    purity  normFactor  score   diploidProportion  ploidy  gender  status  polyclonalProportion  minPurity  maxPurity \
    minPloidy  maxPloidy  minDiploidProportion  maxDiploidProportion  version  somaticPenalty  wholeGenomeDuplication \
    msIndelsPerMb         msStatus  tml  tmlStatus  tmbPerMb            tmbStatus
    0.6300  1.1600      0.5126  0.0000             2.0036  MALE    NORMAL  0.0000                0.6000     0.6400 \
    1.9480     2.0037     0.0000                0.0000                2.40     0.0000          false \
    0.012941587967820916  MSS       0.0  LOW        1.0514165792235046  LOW
    """

    header, values = [], []
    for i, line in enumerate(f["f"].splitlines()):
        fields = line.strip().split("\t")
        if i == 0:
            header = fields
        else:
            values = fields

    return dict(zip(header, values))



""" MultiQC module to parse output from Peddy """


import json
import logging
import random

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import scatter

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    Peddy module class, parses stderr logs.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Peddy",
            anchor="peddy",
            href="https://github.com/brentp/peddy",
            info="calculates genotype :: pedigree correspondence checks, ancestry checks and sex checks using VCF files.",
            doi="10.1016/j.ajhg.2017.01.017",
        )

        # Find and load any Peddy reports
        self.peddy_data = dict()
        self.peddy_length_counts = dict()
        self.peddy_length_exp = dict()
        self.peddy_length_obsexp = dict()

        # parse peddy summary file
        for f in self.find_log_files("peddy/summary_table"):
            parsed_data = self.parse_peddy_summary(f)
            if parsed_data is not None:
                for s_name in parsed_data:
                    cleaned_s_name = self.clean_s_name(s_name, f)
                    try:
                        self.peddy_data[cleaned_s_name].update(parsed_data[s_name])
                    except KeyError:
                        self.peddy_data[cleaned_s_name] = parsed_data[s_name]
                self.add_data_source(f)

                # Superfluous function call to confirm that it is used in this module
                # Replace None with actual version if it is available
                self.add_software_version(None, cleaned_s_name)

        # parse peddy CSV files
        for pattern in ["het_check", "ped_check", "sex_check"]:
            sp_key = f"peddy/{pattern}"
            for f in self.find_log_files(sp_key):
                # some columns have the same name in het_check and sex_check (median_depth)
                # pass pattern to parse_peddy_csv so the column names can include pattern to
                # avoid being overwritten
                parsed_data = self.parse_peddy_csv(f, pattern)
                if parsed_data is not None:
                    for s_name in parsed_data:
                        try:
                            self.peddy_data[s_name].update(parsed_data[s_name])
                        except KeyError:
                            self.peddy_data[s_name] = parsed_data[s_name]

        # parse background PCA JSON file, this is identical for all peddy runs,
        # so just parse the first one we find
        for f in self.find_log_files("peddy/background_pca"):
            background = json.loads(f["f"])
            PC1 = [x["PC1"] for x in background]
            PC2 = [x["PC2"] for x in background]
            ancestry = [x["ancestry"] for x in background]
            self.peddy_data["background_pca"] = {"PC1": PC1, "PC2": PC2, "ancestry": ancestry}
            break

        # Filter to strip out ignored sample names
        self.peddy_data = self.ignore_samples(self.peddy_data)

        if len(self.peddy_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.peddy_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.peddy_data, "multiqc_peddy")

        # Basic Stats Table
        self.peddy_general_stats_table()

        # PCA plot
        self.peddy_pca_plot()

        # Relatedness plot
        self.peddy_relatedness_plot()

        # hetcheck plot
        self.peddy_het_check_plot()

        self.peddy_sex_check_plot()

    @staticmethod
    def parse_peddy_summary(f):
        """Go through log file looking for peddy output"""
        parsed_data = dict()
        headers = None
        for line in f["f"].splitlines():
            s = line.split("\t")
            if headers is None:
                s[0] = s[0].lstrip("#")
                headers = s
            else:
                parsed_data[s[1]] = dict()
                for i, v in enumerate(s):
                    if i != 1:
                        try:
                            parsed_data[s[1]][headers[i]] = float(v)
                        except ValueError:
                            parsed_data[s[1]][headers[i]] = v
        if len(parsed_data) == 0:
            return None
        return parsed_data

    def parse_peddy_csv(self, f, pattern):
        """Parse csv output from peddy"""
        parsed_data = dict()
        headers = None
        s_name_idx = None
        for line in f["f"].splitlines():
            s = line.split(",")
            if headers is None:
                headers = s
                try:
                    s_name_idx = [headers.index("sample_id")]
                except ValueError:
                    try:
                        s_name_idx = [headers.index("sample_a"), headers.index("sample_b")]
                    except ValueError:
                        log.warning(f"Could not find sample name in Peddy output: {f['fn']}")
                        return None
            else:
                s_name = "-".join([s[idx] for idx in s_name_idx])
                s_name = self.clean_s_name(s_name, f)
                parsed_data[s_name] = dict()
                for i, v in enumerate(s):
                    if i not in s_name_idx:
                        if headers[i] == "error" and pattern == "sex_check":
                            v = "True" if v == "False" else "False"
                        try:
                            # add the pattern as a suffix to key
                            parsed_data[s_name][headers[i] + "_" + pattern] = float(v)
                        except ValueError:
                            # add the pattern as a suffix to key
                            parsed_data[s_name][headers[i] + "_" + pattern] = v
        if len(parsed_data) == 0:
            return None
        return parsed_data

    def peddy_general_stats_table(self):
        """Take the parsed stats from the Peddy report and add it to the
        basic stats table at the top of the report"""

        family_ids = [x.get("family_id") for x in self.peddy_data.values()]

        headers = {
            "family_id": {
                "title": "Family ID",
                "hidden": True if all([v == family_ids[0] for v in family_ids]) else False,
            },
            "ancestry-prediction": {
                "title": "Ancestry",
                "description": "Ancestry Prediction",
            },
            "ancestry-prob_het_check": {
                "title": "P(Ancestry)",
                "description": "Probability predicted ancestry is correct.",
            },
            "sex_het_ratio": {
                "title": "Sex / Het Ratio",
            },
            "error_sex_check": {
                "title": "Correct Sex",
                "description": "Displays False if error in sample sex prediction",
            },
            "predicted_sex_sex_check": {"title": "Sex", "description": "Predicted sex"},
        }
        self.general_stats_addcols(self.peddy_data, headers)

    def peddy_pca_plot(self):
        ancestry_colors = {
            "SAS": "68,1,81",
            "EAS": "59,81,139",
            "AMR": "33,144,141",
            "AFR": "92,200,99",
            "EUR": "253,231,37",
        }
        default_color = "0,0,0"
        default_background_color = "211,211,211"

        data = {}

        # plot the background data first, so it doesn't hide the actual data points
        d = self.peddy_data.pop("background_pca", {})
        if d:
            background = [
                {
                    "x": pc1,
                    "y": pc2,
                    "color": f"rgba({ancestry_colors.get(ancestry, default_background_color)},0.1)",
                    "name": ancestry,
                    "marker_size": 3,
                    "marker_line_width": 0,
                    "annotate": False,
                }
                for pc1, pc2, ancestry in zip(
                    d["PC1"],
                    d["PC2"],
                    d["ancestry"],
                )
            ]
            data["background"] = background

        for s_name, d in self.peddy_data.items():
            if "PC1_het_check" in d and "PC2_het_check" in d:
                data[s_name] = {
                    "x": d["PC1_het_check"],
                    "y": d["PC2_het_check"],
                }
                try:
                    data[s_name]["color"] = f"rgb({ancestry_colors.get(d['ancestry-prediction'], default_color)})"
                except KeyError:
                    pass

        pconfig = {
            "id": "peddy_pca_plot",
            "title": "Peddy: PCA Plot",
            "xlab": "PC1",
            "ylab": "PC2",
        }

        if len(data) > 0:
            self.add_section(
                name="PCA Plot",
                anchor="peddy-pca-plot",
                plot=scatter.plot(data, pconfig),
            )

    def peddy_relatedness_plot(self):
        data = dict()
        for s_name, d in self.peddy_data.items():
            if "ibs0_ped_check" in d and "ibs2_ped_check" in d:
                data[s_name] = {"x": d["ibs0_ped_check"], "y": d["ibs2_ped_check"]}
            if "rel_ped_check" in d:
                if d["rel_ped_check"] < 0.25:
                    data[s_name]["color"] = "rgba(109, 164, 202, 0.9)"
                elif d["rel_ped_check"] < 0.5:
                    data[s_name]["color"] = "rgba(250, 160, 81, 0.8)"
                else:
                    data[s_name]["color"] = "rgba(43, 159, 43, 0.8)"

        pconfig = {
            "id": "peddy_relatedness_plot",
            "title": "Peddy: Relatedness Plot",
            "xlab": "IBS0 (no alleles shared)",
            "ylab": "IBS2 (both alleles shared)",
        }

        if len(data) > 0:
            self.add_section(
                name="Relatedness",
                anchor="peddy-relatedness-plot",
                description="""Shared allele rates between sample pairs. Points are coloured by degree of relatedness:
                <span style="color: #6DA4CA;">less than 0.25</span>,
                <span style="color: #FAA051;">0.25 - 0.5</span>,
                <span style="color: #2B9F2B;">greather than 0.5</span>.""",
                plot=scatter.plot(data, pconfig),
            )

    def peddy_het_check_plot(self):
        """plot the het_check scatter plot"""
        # empty dictionary to add sample names, and dictionary of values
        data = {}

        # for each sample, and list in self.peddy_data
        for s_name, d in self.peddy_data.items():
            # check the sample contains the required columns
            if "median_depth_het_check" in d and "het_ratio_het_check" in d:
                # add sample to dictionary with value as a dictionary of points to plot
                data[s_name] = {
                    "x": d["median_depth_het_check"],
                    "y": d["het_ratio_het_check"],
                }

        pconfig = {
            "id": "peddy_het_check_plot",
            "title": "Peddy: Het Check",
            "xlab": "median depth",
            "ylab": "proportion het calls",
        }

        if len(data) > 0:
            self.add_section(
                name="Het Check",
                description="Proportion of sites that were heterozygous against median depth.",
                helptext="""
                A high proportion of heterozygous sites suggests contamination, a low proportion suggests consanguinity.

                See [the main peddy documentation](https://peddy.readthedocs.io/en/latest/output.html#het-check) for more details about the `het_check` command.
                """,
                anchor="peddy-hetcheck-plot",
                plot=scatter.plot(data, pconfig),
            )

    def peddy_sex_check_plot(self):
        data = {}
        sex_index = {"female": 0, "male": 1, "unknown": 2}

        for s_name, d in self.peddy_data.items():
            if "sex_het_ratio" in d and "ped_sex_sex_check" in d:
                data[s_name] = {
                    "x": sex_index.get(d["ped_sex_sex_check"], 2) + (random.random() - 0.5) * 0.1,
                    "y": d["sex_het_ratio"],
                }

        pconfig = {
            "id": "peddy_sex_check_plot",
            "title": "Peddy: Sex Check",
            "xlab": "Sex From Ped",
            "ylab": "Sex Het Ratio",
            "categories": ["Female", "Male", "Unknown"],
        }

        if len(data) > 0:
            self.add_section(
                name="Sex Check",
                description="Predicted sex against heterozygosity ratio",
                helptext="""
                Higher values of Sex Het Ratio suggests the sample is female, low values suggest male.

                See [the main peddy documentation](http://peddy.readthedocs.io/en/latest/#sex-check) for more details about the `het_check` command.
                """,
                anchor="peddy-sexcheck-plot",
                plot=scatter.plot(data, pconfig),
            )



""" MultiQC submodule to parse output from deepTools plotPCA """

import logging

from multiqc.plots import scatter

# Initialise the logger
log = logging.getLogger(__name__)


class plotPCAMixin:
    def parse_plotPCA(self):
        """Find plotPCA output"""
        self.deeptools_plotPCAData = dict()
        for f in self.find_log_files("deeptools/plotPCAData", filehandles=False):
            parsed_data = self.parsePlotPCAData(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_plotPCAData:
                    log.warning(f"Replacing duplicate sample {k}.")
                self.deeptools_plotPCAData[k] = v
            if len(parsed_data) > 0:
                self.add_data_source(f, section="plotPCA")

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        self.deeptools_plotPCAData = self.ignore_samples(self.deeptools_plotPCAData)

        if len(self.deeptools_plotPCAData) > 0:
            # Write data to file
            self.write_data_file(self.deeptools_plotPCAData, "deeptools_plot_PCA")

            config = {
                "id": "deeptools_pca_plot",
                "title": "deeptools: PCA Plot",
                "xlab": "PC1",
                "ylab": "PC2",
                "tt_label": "PC1 {point.x:.2f}: PC2 {point.y:.2f}",
            }
            data = dict()
            for s_name in self.deeptools_plotPCAData:
                try:
                    data[s_name] = {
                        "x": self.deeptools_plotPCAData[s_name][1],
                        "y": self.deeptools_plotPCAData[s_name][2],
                    }
                except KeyError:
                    pass
            if len(data) == 0:
                log.debug("No valid data for PCA plot")
                return None

            self.add_section(
                name="PCA plot",
                anchor="deeptools_pca",
                description="PCA plot with the top two principal components calculated based on genome-wide distribution of sequence reads",
                plot=scatter.plot(data, config),
            )

        return len(self.deeptools_plotPCAData)

    def parsePlotPCAData(self, f):
        d = dict()
        samples = []
        for line in f["f"].splitlines():
            cols = line.strip().split("\t")
            if cols[0] == "#plotPCA --outFileNameData":
                continue
            elif cols[0] == "Component":
                for c in cols[1 : (len(cols) - 1)]:
                    c = str(c).strip("'")
                    s_name = self.clean_s_name(c, f)
                    d[s_name] = {}
                    samples.append(s_name)
            else:
                idx = 0
                compo = cols[0]
                for c in cols[1 : (len(cols) - 1)]:
                    d[samples[idx]][self._int(compo)] = float(c)
                    idx += 1
        return d



""" MultiQC module to parse output from Preseq """


import logging

import numpy as np

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph
from multiqc.utils import mqc_colour

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Preseq",
            anchor="preseq",
            href="http://smithlabresearch.org/software/preseq/",
            info="""estimates the complexity of a library, showing how many additional
                    unique reads are sequenced for increasing total read count.
                    A shallow curve indicates complexity saturation. The dashed line
                    shows a perfectly complex library where total reads = unique reads.""",
            doi="10.1038/nmeth.2375",
        )

        # Find and load any Preseq reports
        reads_data = dict()
        bases_data = dict()
        for f in self.find_log_files("preseq"):
            sample_data_raw, sample_data_is_bases = _parse_preseq_logs(f)
            if sample_data_raw is None:
                continue

            if f["s_name"] in sample_data_raw:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")

            if sample_data_is_bases:
                bases_data[f["s_name"]] = sample_data_raw
            else:
                reads_data[f["s_name"]] = sample_data_raw

            self.add_data_source(f)

        # Filter to strip out ignored sample names
        bases_data = self.ignore_samples(bases_data)
        reads_data = self.ignore_samples(reads_data)
        all_data = {**bases_data, **reads_data}
        if not all_data:
            raise ModuleNoSamplesFound

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        if bases_data and reads_data:
            log.warning("Mixed 'TOTAL_READS' and 'TOTAL_BASES' reports. Will build two separate plots")
        log.info(f"Found {len(all_data)} reports")
        # Write data to file
        self.write_data_file(all_data, "preseq")

        # Preseq plot. If mixed logs are found, build two separate plots.
        if bases_data:
            self._make_preseq_length_trimmed_plot(bases_data, True)
        if reads_data:
            self._make_preseq_length_trimmed_plot(reads_data, False)

    def _make_preseq_length_trimmed_plot(self, data_raw, is_basepairs):
        """Generate the preseq plot.

        For Y axis, plot coverages if `config.preseq.read_length` and `config.preseq.genome_size`
        specified (unless requested otherwise in `config.preseq.y_axis`).

        For X axis, plot counts (or base pairs) in X axis (unless coverages requested
        explicitly in `config.preseq.x_axis`).
        """
        counts_in_1x = _get_counts_in_1x(is_basepairs)
        x_axis = getattr(config, "preseq", {}).get("x_axis", "counts")
        y_axis = getattr(config, "preseq", {}).get("y_axis", "coverage" if counts_in_1x else "counts")

        # Modify counts
        d_cnts = {sn: _modify_raw_data(sample_data, is_basepairs) for sn, sample_data in data_raw.items()}

        # Convert counts (base pairs) -> depths
        d_covs = {
            sn: _counts_to_coverages(sample_data, counts_in_1x) if counts_in_1x else None
            for sn, sample_data in data_raw.items()
        }

        # Prepare final dataset for plotting
        data = dict()
        for sn, s_d_cnts, s_d_covs in zip(d_cnts, d_cnts.values(), d_covs.values()):
            keys = s_d_covs.keys() if x_axis == "coverage" else s_d_cnts.keys()
            values = s_d_covs.values() if y_axis == "coverage" else s_d_cnts.values()
            data[sn] = dict(zip(keys, values))

        # Count maximum values to draw the "ideal" line
        max_y_raw, max_sn = max((max(sd.values()), sn) for sn, sd in data_raw.items())
        max_y_cnt = list(_modify_raw_data({max_y_raw: max_y_raw}, is_basepairs).items())[0][0]
        max_y_cov = list(_counts_to_coverages({max_y_raw: max_y_raw}, counts_in_1x).items())[0][0]
        max_y = max_y_cov if y_axis == "coverage" else max_y_cnt
        max_yx = max_y_cov if x_axis == "coverage" else max_y_cnt

        # Preparing axis and tooltip labels
        x_suffix, y_tt_lbl, x_axis_name, y_suffix, x_tt_lbl, y_axis_name = _prepare_labels(
            is_basepairs, max_y_cov, x_axis, y_axis
        )

        name = "Complexity curve"
        description = ""
        section_id = "preseq_plot"
        pconfig = {
            "id": "preseq_complexity_plot",
            "title": "Preseq: Complexity curve",
            "xlab": x_axis_name,
            "ylab": y_axis_name,
            "xmin": 0,
            "ymin": 0,
            "tt_label": "<b>" + y_tt_lbl + "</b>: " + x_tt_lbl,
            "xsuffix": x_suffix,
            "ysuffix": y_suffix,
            "extra_series": [],
        }
        if not is_basepairs:
            pconfig["title"] += " (molecule count)"
            pconfig["id"] += "_molecules"
            name += " (molecule count)"
            section_id += "_molecules"

        # Parse the real counts if we have them
        real_cnts_all, real_cnts_unq = self._parse_real_counts(data.keys())
        real_vals_all, real_vals_unq = _prep_real_counts(
            real_cnts_all, real_cnts_unq, is_basepairs, counts_in_1x, x_axis, y_axis
        )
        pconfig["extra_series"].extend(
            _real_counts_to_plot_series(data, real_vals_unq, real_vals_all, x_suffix, y_suffix, y_tt_lbl)
        )
        if real_vals_unq:
            description += "<p>Points show read count versus deduplicated read counts (externally calculated).</p>"
        elif real_vals_all:
            description += "<p>Points show externally calculated read counts on the curves.</p>"

        # Trim the data to not have a ridiculous x-axis (10Gbp anyone?)
        if getattr(config, "preseq", {}).get("notrim", False) is not True:
            max_y *= 0.8
            max_yx *= 0.8
            max_x = 0
            for x in sorted(list(data[max_sn].keys())):
                max_x = max(max_x, x)
                if data[max_sn][x] > max_y and x > real_vals_all.get(max_sn, 0) and x > real_vals_unq.get(max_sn, 0):
                    break
            pconfig["xmax"] = max_x
            description += "<p>Note that the x-axis is trimmed at the point where all the datasets \
                show 80% of their maximum y-value, to avoid ridiculous scales.</p>"

        # Plot perfect library as dashed line
        pconfig["extra_series"].append(
            {
                "name": "a perfect library where each read is unique",
                "data": [[0, 0], [max_yx, max_y]],
                "dashStyle": "Dash",
                "lineWidth": 1,
                "color": "#000000",
                "marker": {"enabled": False},
                "showInLegend": False,
            }
        )

        self.add_section(name=name, description=description, anchor=section_id, plot=linegraph.plot(data, pconfig))

    def _parse_real_counts(self, sample_names):
        real_counts_file_raw = None
        real_counts_file_name = None
        for f in self.find_log_files("preseq/real_counts"):
            if real_counts_file_raw is not None:
                log.warning(f"Multiple Preseq real counts files found, now using {f['fn']}")
            real_counts_file_raw = f["f"]
            real_counts_file_name = f["fn"]

        real_counts_total = {}
        real_counts_unique = {}

        if real_counts_file_raw is not None:
            try:
                for line in real_counts_file_raw.splitlines():
                    if not line.startswith("#"):
                        cols = line.strip().split()  # Split on any whitespace
                        sn = self.clean_s_name(cols[0], None)
                        if sn in sample_names:
                            if len(cols) >= 2:
                                if cols[1].isdigit():
                                    real_counts_total[sn] = int(cols[1])
                            if len(cols) >= 3:
                                if cols[2].isdigit():
                                    real_counts_unique[sn] = int(cols[2])
            except IOError as e:
                log.error(f"Error loading real counts file {real_counts_file_name}: {str(e)}")
            else:
                log.debug(f"Found {len(real_counts_total)} matching sets of counts from {real_counts_file_name}")

        return real_counts_total, real_counts_unique


def _parse_preseq_logs(f):
    """Go through log file looking for preseq output"""

    lines = f["f"].splitlines()
    header = lines.pop(0)

    data_is_bases = False
    if header.startswith("TOTAL_READS	EXPECTED_DISTINCT"):
        pass
    elif header.startswith("TOTAL_BASES	EXPECTED_DISTINCT"):
        data_is_bases = True
    elif header.startswith("total_reads	distinct_reads"):
        pass
    else:
        log.debug(f"First line of preseq file {f['fn']} did not look right")
        return None, None

    data = dict()
    for line in lines:
        s = line.split()
        # Sometimes the Expected_distinct count drops to 0, not helpful
        if float(s[1]) == 0 and float(s[0]) > 0:
            continue
        data[float(s[0])] = float(s[1])

    return data, data_is_bases


def _modify_raw_data(sample_data, is_basepairs):
    """Modify counts or base pairs according to `read_count_multiplier`
    or `base_count_multiplier`.
    """
    return {_modify_raw_val(x, is_basepairs): _modify_raw_val(y, is_basepairs) for x, y in sample_data.items()}


def _modify_raw_val(val, is_basepairs):
    """Modify counts or base pairs according to `read_count_multiplier`
    or `base_count_multiplier`.
    """
    return float(val) * (config.base_count_multiplier if is_basepairs else config.read_count_multiplier)


def _counts_to_coverages(sample_data, counts_in_1x):
    """If the user specified read length and genome size in the config,
    convert the raw counts/bases into the depth of coverage.
    """
    if not counts_in_1x:
        return {None: None}

    return {_count_to_coverage(x, counts_in_1x): _count_to_coverage(y, counts_in_1x) for x, y in sample_data.items()}


def _count_to_coverage(val, counts_in_1x):
    return val / counts_in_1x


def _get_counts_in_1x(data_is_basepairs):
    """Read length and genome size from the config and calculate
    the approximate number of counts (or base pairs) in 1x of depth
    """
    read_length = float(getattr(config, "preseq", {}).get("read_length", 0))
    genome_size = getattr(config, "preseq", {}).get("genome_size")
    if genome_size:
        try:
            genome_size = float(genome_size)
        except ValueError:
            presets = {"hg19_genome": 2897310462, "hg38_genome": 3049315783, "mm10_genome": 2652783500}
            if genome_size in presets:
                genome_size = presets[genome_size]
            else:
                log.warning(
                    "The size for genome "
                    + genome_size
                    + " is unknown to MultiQC, "
                    + "please specify it explicitly or choose one of the following: "
                    + ", ".join(presets.keys())
                    + ". Falling back to molecule counts."
                )
                genome_size = None

    if genome_size:
        if data_is_basepairs:
            return genome_size
        elif read_length:
            return genome_size / read_length
    else:
        return None


def _prepare_labels(is_basepairs, max_y_cov, x_axis, y_axis):
    cov_suffix = "x"

    cov_lbl = None
    if x_axis == "coverage" or y_axis == "coverage":
        cov_precision = "2"
        if max_y_cov > 30:  # no need to be so precise when the depth numbers are high
            cov_precision = "1"
        if max_y_cov > 300:  # when the depth are very high, decimal digits are excessive
            cov_precision = "0"
        cov_lbl = "{value:,." + cov_precision + "f}x"

    cnt_lbl = "{value:,.2f} " + config.read_count_prefix
    cnt_suffix = config.read_count_prefix
    if config.read_count_multiplier == 1:
        cnt_lbl = "{value:,.0f}"
    if is_basepairs:
        cnt_lbl = "{value:,.2f} " + config.base_count_prefix
        cnt_suffix = " " + config.base_count_prefix
        if config.base_count_multiplier == 1:
            cnt_lbl = "{value:,.0f}"

    if x_axis == "coverage":
        x_suffix = cov_suffix
        x_tt_lbl = cov_lbl.replace("value", "point.x")
        x_axis_name = "Total coverage (including duplicates)"
    else:
        x_suffix = cnt_suffix
        x_tt_lbl = cnt_lbl.replace("value", "point.x")
        if is_basepairs:
            x_tt_lbl += " pairs (total)"
            x_axis_name = "Base pairs (total)"
        else:
            x_tt_lbl += " total molecules"
            x_axis_name = "Total molecules (including duplicates)"

    if y_axis == "coverage":
        y_suffix = cov_suffix
        y_tt_lbl = cov_lbl.replace("value", "point.y") + " depth"
        y_axis_name = "Unique coverage"
    else:
        y_suffix = cnt_suffix
        y_tt_lbl = cnt_lbl.replace("value", "point.y")
        if is_basepairs:
            y_tt_lbl += " pairs (unique)"
            y_axis_name = "Base pairs (unique reads)"
        else:
            y_tt_lbl += " unique molecules"
            y_axis_name = "Unique molecules"
    return x_suffix, y_tt_lbl, x_axis_name, y_suffix, x_tt_lbl, y_axis_name


def _prep_real_counts(real_cnts_all, real_cnts_unq, is_basepairs, counts_in_1x, x_axis, y_axis):
    if x_axis == "coverage":
        real_vals_all = {sn: _count_to_coverage(x, counts_in_1x) for sn, x in real_cnts_all.items()}
    else:
        real_vals_all = {sn: _modify_raw_val(x, is_basepairs) for sn, x in real_cnts_all.items()}

    if y_axis == "coverage":
        real_vals_unq = {sn: _count_to_coverage(y, counts_in_1x) for sn, y in real_cnts_unq.items()}
    else:
        real_vals_unq = {sn: _modify_raw_val(y, is_basepairs) for sn, y in real_cnts_unq.items()}

    return real_vals_all, real_vals_unq


def _real_counts_to_plot_series(data, yx_by_sample, xs_by_sample, x_suffix, y_suffix, y_tt_lbl):
    scale = mqc_colour.mqc_colour_scale("plot_defaults")

    series = []
    for si, sn in enumerate(sorted(data.keys())):
        if sn in xs_by_sample:
            x = float(xs_by_sample[sn])
            point = {
                "color": scale.get_colour(si),
                "showInLegend": False,
                "marker": {
                    "enabled": True,
                    "symbol": "diamond",
                    "lineColor": "black",
                    "lineWidth": 1,
                },
            }
            if sn in yx_by_sample:
                y = float(yx_by_sample[sn])
                point["data"] = [[x, y]]
                point["name"] = f"{sn}: actual read count vs. deduplicated read count (externally calculated)"
                series.append(point)
                y = y_tt_lbl.replace("point.y", "y").format(y=y)
                log.debug(f"Real counts for {sn}: {x}{x_suffix}, ({y}{y_suffix})")
            else:
                xs = sorted(data[sn].keys())
                ys = sorted(data[sn].values())
                if x > max(xs):
                    log.warning(
                        f"Total reads for {sn} ({x}{x_suffix}) > max preseq value ({max(xs)}{x_suffix}): "
                        "skipping this point"
                    )
                else:
                    interp_y = np.interp(x, xs, ys)
                    point["data"] = [[x, interp_y]]
                    point["name"] = sn + ": actual read count (externally calculated)"
                    series.append(point)
                    y = y_tt_lbl.replace("point.y", "y").format(y=interp_y)
                    log.debug(f"Real count for {sn}: {x}{{x_suffix}} ({y}{{y_suffix}})")
    return series



""" MultiQC module to parse output from ngs-disambiguate. """

import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Disambiguate",
            anchor="disambiguate",
            href="https://github.com/AstraZeneca-NGS/disambiguate",
            info="is an algorithm for disambiguating reads aligned to two different species (e.g. human and mouse).",
            doi="10.12688/f1000research.10082.1",
        )

        self.data = {}

        for f in self.find_log_files("disambiguate"):
            summ_data = self.parse_summary(f["f"])

            for sample, counts in summ_data.items():
                # Clean sample name.
                sample = self.clean_s_name(sample, f)

                # Check for duplicates.
                if sample in self.data:
                    log.debug("Duplicate sample name found! Overwriting: %s", sample)

                # Add to data and add data source.
                self.data[sample] = counts
                self.add_data_source(f, s_name=sample)

        self.data = self.ignore_samples(self.data)

        if len(self.data) == 0:
            raise ModuleNoSamplesFound

        log.info("Found %d reports", len(self.data))

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.add_stats_table()
        self.add_stats_plot()
        self.write_stats_to_file()

    @staticmethod
    def parse_summary(contents):
        """Parses summary file into a dictionary of counts."""

        lines = contents.strip().split("\n")

        data = {}
        for row in lines[1:]:
            split = row.strip().split("\t")

            sample = split[0]

            data[sample] = {"species_a": int(split[1]), "species_b": int(split[2]), "ambiguous": int(split[3])}

        return data

    def add_stats_table(self):
        """Adds stats to general table."""

        totals = {sample: sum(counts.values()) for sample, counts in self.data.items()}

        percentages = {
            sample: {k: (v / totals[sample]) * 100 for k, v in counts.items()} for sample, counts in self.data.items()
        }

        headers = {
            "species_a": {
                "title": "% Species a",
                "description": "Percentage of reads mapping to species a",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            }
        }

        self.general_stats_addcols(percentages, headers)

    def write_stats_to_file(self):
        """Writes stats to data file."""
        self.write_data_file(self.data, "multiqc_disambiguate")

    def add_stats_plot(self):
        """Plots alignment stats as bargraph."""

        keys = {
            "species_a": {"color": "#437bb1", "name": "Species a"},
            "species_b": {"color": "#b1084c", "name": "Species b"},
            "ambiguous": {"color": "#333333", "name": "Ambiguous"},
        }

        plot_config = {
            "id": "disambiguated_alignments",
            "title": "Disambiguate: Alignment Counts",
            "cpswitch_counts_label": "# Reads",
            "ylab": "# Reads",
        }

        self.add_section(plot=bargraph.plot(self.data, keys, plot_config))



""" MultiQC module to parse output from sargasso """


import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="sargasso",
            anchor="sargasso",
            href="http://biomedicalinformaticsgroup.github.io/Sargasso/",
            info="is a tool to separate mixed-species RNA-seq reads" "according to their species of origin.",
            doi="10.1038/s41596-018-0029-2",
        )

        # Find and load any Sargasso reports
        self.sargasso_data = dict()
        self.sargasso_files = list()
        self.sargasso_keys = list()  # header keys

        for f in self.find_log_files("sargasso"):
            self.parse_sargasso_logs(f)
            self.sargasso_files.append(f)
            self.add_data_source(f)

        # log.info('Removing ignored samples...')
        self.sargasso_data = self.ignore_samples(self.sargasso_data)

        if len(self.sargasso_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.sargasso_files)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.sargasso_data, "multiqc_sargasso")

        # Basic Stats Table
        self.sargasso_stats_table()

        # Assignment bar plot
        self.add_section(plot=self.sargasso_chart())

    def parse_sargasso_logs(self, f):
        """Parse the sargasso log file."""
        species_name = list()
        items = list()
        is_first_line = True
        for line in f["f"].splitlines():
            s = line.split(",")
            # Check that this actually is a Sargasso file
            if is_first_line and s[0] != "Sample":
                return None

            if len(s) < 7:
                continue
            if is_first_line:
                # prepare header
                is_first_line = False
                header = s
                for i in header[1:]:
                    # find out what species included
                    sname = i.split("-")[-1]
                    if sname not in species_name:
                        species_name.append(sname)
                    # find out what is being counted
                    kname = "-".join(i.split("-")[-3:-1])
                    if kname not in items:
                        items.append(kname)
            else:
                # start sample lines.
                sample_name = s.pop(0)

                chunk_by_species = [s[i : i + len(items)] for i in range(0, len(s), len(items))]
                for idx, v in enumerate(chunk_by_species):
                    # adding species name to the sample name for easy interpretation
                    new_sample_name = "_".join([sample_name, species_name[idx]])

                    # Clean up sample name
                    new_sample_name = self.clean_s_name(new_sample_name, f)

                    if new_sample_name in self.sargasso_data.keys():
                        log.debug(f"Duplicate sample name found! Overwriting: {new_sample_name}")

                    try:
                        self.sargasso_data[new_sample_name] = dict(zip(items, map(int, v)))
                    except ValueError:
                        pass

        self.sargasso_keys = items

        for idx, f_name in enumerate(self.sargasso_data.keys()):
            # Reorganised parsed data for this sample
            # Collect total READ count number
            self.sargasso_data[f_name]["Total"] = 0
            for key, value in list(self.sargasso_data[f_name].items()):  # iter on both keys and values
                if key.endswith("Reads"):
                    self.sargasso_data[f_name]["Total"] += value

            # Calculate the percent aligned if we can
            try:
                self.sargasso_data[f_name]["sargasso_percent_assigned"] = (
                    float(self.sargasso_data[f_name]["Assigned-Reads"]) / float(self.sargasso_data[f_name]["Total"])
                ) * 100.0
            except (KeyError, ZeroDivisionError):
                pass

    def sargasso_stats_table(self):
        """Take the parsed stats from the sargasso report and add them to the
        basic stats table at the top of the report"""

        headers = {
            "sargasso_percent_assigned": {
                "title": "% Assigned",
                "description": "Sargasso % Assigned reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn",
            },
            "Assigned-Reads": {
                "title": f"{config.read_count_prefix} Assigned",
                "description": f"Sargasso Assigned reads ({config.read_count_desc})",
                "min": 0,
                "scale": "PuBu",
                "modify": lambda x: float(x) * config.read_count_multiplier,
                "shared_key": "read_count",
            },
        }
        self.general_stats_addcols(self.sargasso_data, headers)

    def sargasso_chart(self):
        """Make the sargasso plot"""

        # Config for the plot
        config = {
            "id": "sargasso_assignment_plot",
            "title": "Sargasso: Assigned Reads",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        # We only want to plot the READs at the moment
        return bargraph.plot(self.sargasso_data, [name for name in self.sargasso_keys if "Reads" in name], config)



""" MultiQC module to parse output from UMI-tools """


import logging
import re
from typing import Dict, Optional

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, beeswarm

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    umitools module class, parses dedup logs
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="UMI-tools",
            anchor="umitools",
            href="https://github.com/CGATOxford/UMI-tools",
            info="contains tools for dealing with Unique Molecular Identifiers (UMIs)/(RMTs) and scRNA-Seq barcodes.",
            doi="10.1101/gr.209601.116",
        )

        dedup_data_by_sample = dict()
        for f in self.find_log_files("umitools/dedup"):
            # Parse the log file for sample name and statistics
            data = self.parse_dedup_logs(f)
            if data:
                f["s_name"] = self._parse_s_name(f) or f["s_name"]
                if f["s_name"] in dedup_data_by_sample:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                dedup_data_by_sample[f["s_name"]] = data
                self.add_data_source(f)  # Add the log file information to the multiqc_sources.txt
                if "version" in data:  # Add version info
                    self.add_software_version(data.pop("version"), f["s_name"])

        extract_data_by_sample = dict()
        for f in self.find_log_files("umitools/extract"):
            # Parse the log file for sample name and statistics
            data = self.parse_extract_logs(f)
            if data:
                f["s_name"] = self._parse_s_name(f) or f["s_name"]
                if f["s_name"] in extract_data_by_sample:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                extract_data_by_sample[f["s_name"]] = data
                self.add_data_source(f)  # Add the log file information to the multiqc_sources.txt
                if "version" in data:  # Add version info
                    self.add_software_version(data.pop("version"), f["s_name"])

        # Check the log files against the user supplied list of samples to ignore
        dedup_data_by_sample = self.ignore_samples(dedup_data_by_sample)
        extract_data_by_sample = self.ignore_samples(extract_data_by_sample)
        if len(dedup_data_by_sample) == 0 and len(extract_data_by_sample) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(dedup_data_by_sample)} deduplication reports")
        log.info(f"Found {len(extract_data_by_sample)} extract reports")

        # Write parsed reports data to a file
        self.write_data_file(dedup_data_by_sample, "multiqc_umitools_dedup")
        self.write_data_file(extract_data_by_sample, "multiqc_umitools_extract")

        if dedup_data_by_sample:
            self.dedup_general_stats_table(dedup_data_by_sample)
            self.deduplication_plot(dedup_data_by_sample)
            self.umi_stats_violin(dedup_data_by_sample)

        if extract_data_by_sample:
            self.extract_general_stats_table(extract_data_by_sample)
            self.umitools_extract_barplot(extract_data_by_sample)

    def _parse_s_name(self, f) -> Optional[str]:
        # Get the s_name from the input file if possible
        # stdin : <_io.TextIOWrapper name='M18-39155_T1.Aligned.sortedByCoord.out.bam' mode='r' encoding='UTF-8'>
        s_name_re = r"stdin\s+:\s+<_io\.TextIOWrapper name='([^\']+)'"
        s_name_match = re.search(s_name_re, f["f"])
        if s_name_match:
            return self.clean_s_name(s_name_match.group(1))
        return None

    @staticmethod
    def parse_dedup_logs(f) -> Dict:
        regexes = [
            (int, "total_umis", r"INFO total_umis (\d+)"),
            (int, "unique_umis", r"INFO #umis (\d+)"),
            (int, "dedup_input_reads", r"INFO Reads: Input Reads: (\d+)"),
            (int, "dedup_output_reads", r"INFO Number of reads out: (\d+)"),
            (int, "positions_deduplicated", r"INFO Total number of positions deduplicated: (\d+)"),
            (float, "mean_umi_per_pos", r"INFO Mean number of unique UMIs per position: ([\d\.]+)"),
            (float, "max_umi_per_pos", r"INFO Max. number of unique UMIs per position: (\d+)"),
            (str, "version", r"# UMI-tools version: ([\d\.]+)"),
        ]

        data = {}
        # Search for values using regular expressions
        for type_, key, regex in regexes:
            re_matches = re.search(regex, f["f"])
            if re_matches:
                if key == "version":
                    data[key] = re_matches.group(1)
                else:
                    data[key] = type_(re_matches.group(1))

            # Calculate a few simple supplementary stats
            try:
                data["dedup_percent_passing"] = round(
                    ((data["dedup_output_reads"] / data["dedup_input_reads"]) * 100.0), 2
                )
            except (KeyError, ZeroDivisionError):
                pass
            try:
                data["dedup_removed_reads"] = data["dedup_input_reads"] - data["dedup_output_reads"]
            except KeyError:
                pass

        return data

    @staticmethod
    def parse_extract_logs(f) -> Dict:
        regexes = [
            (int, "extract_input_reads", r"INFO Input Reads: (\d+)"),
            (int, "read1_mismatch", r"INFO regex does not match read1: (\d+)"),
            (int, "read1_match", r"INFO regex matches read1: (\d+)"),
            (int, "read2_mismatch", r"INFO regex does not match read2: (\d+)"),
            (int, "read2_match", r"INFO regex matches read2: (\d+)"),
            (int, "extract_output_reads", r"INFO Reads output: (\d+)"),
            (str, "version", r"# UMI-tools version: ([\d\.]+)"),
        ]

        data = {}
        # Search for values using regular expressions
        for type_, key, regex in regexes:
            re_matches = re.search(regex, f["f"])
            if re_matches:
                if key == "version":
                    data[key] = re_matches.group(1)
                else:
                    data[key] = type_(re_matches.group(1))

        return data

    def dedup_general_stats_table(self, data_by_sample):
        """
        Take the parsed stats from the umitools report and add it to the
        basic stats table at the top of the report
        """
        headers = {
            "dedup_output_reads": {
                "title": f"{config.read_count_prefix} Unique Reads",
                "description": f"Reads remaining after deduplication ({config.read_count_desc})",
                "min": 0,
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "scale": "PuRd",
            },
            "dedup_percent_passing": {
                "title": "% Pass Dedup",
                "description": "% processed reads that passed deduplication",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn",
            },
        }
        self.general_stats_addcols(data_by_sample, headers)

    def extract_general_stats_table(self, data_by_sample):
        headers = {
            "extract_output_reads": {
                "title": f"{config.read_count_prefix} Extracted Reads",
                "description": f"Reads remaining after extraction ({config.read_count_desc})",
                "min": 0,
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
                "scale": "PuRd",
            },
        }
        self.general_stats_addcols(data_by_sample, headers)

    def deduplication_plot(self, data_by_sample):
        """Generate a plot the read deduplication rates for the main report"""

        # Specify the order of the different possible categories
        keys = {
            "dedup_output_reads": {"color": "#7fc97f", "name": "Reads remaining"},
            "dedup_removed_reads": {"color": "#fdc086", "name": "Reads removed"},
        }

        return self.add_section(
            name="Deduplicated Reads",
            anchor="umitools-dedup-plot",
            plot=bargraph.plot(
                data_by_sample,
                keys,
                {
                    "id": "umitools_deduplication_barplot",
                    "title": "UMI-tools: Deduplication Counts",
                    "ylab": "# Reads",
                    "cpswitch_counts_label": "Number of Reads",
                },
            ),
        )

    def umitools_extract_barplot(self, data_by_sample):
        keys = {
            "read1_match": {"color": "#7fc9c7", "name": "Read1 Match"},
            "read1_mismatch": {"color": "#fd9286", "name": "Read1 Mismatch"},
            "read2_match": {"color": "#7f89c9", "name": "Read2 Match"},
            "read2_mismatch": {"color": "#fd86aa", "name": "Read2 Mismatch"},
        }

        # add a section with a beeswarm plot of UMI stats to the report
        self.add_section(
            name="Extract Stats",
            anchor="umitools_extract",
            description="Read stats from `umi_tools extract`",
            plot=bargraph.plot(
                data_by_sample,
                keys,
                {
                    "id": "umitools_extract_barplot",
                    "title": "UMI-tools: Extract Stats",
                    "ylab": "# Reads",
                    "cpswitch_counts_label": "Number of Reads",
                },
            ),
        )

    def umi_stats_violin(self, data_by_sample):
        """Generate a violin of UMI stats for the main report"""

        headers = {
            "positions_deduplicated": {
                "title": "Positions Dedup",
                "description": "Total number of positions deduplicated",
                "min": 0,
                "format": "{:,d}",
                "scale": "Greens",
            },
            "total_umis": {
                "title": "Total UMIs",
                "description": "Total UMIs found in sample",
                "min": 0,
                "format": "{:,d}",
                "scale": "Blues",
            },
            "unique_umis": {
                "title": "Unique UMIs",
                "description": "Unique UMIs found in sample",
                "min": 0,
                "format": "{:,d}",
                "scale": "Purples",
            },
            "mean_umi_per_pos": {
                "title": "Mean #UMI",
                "description": "Mean number of unique UMIs per position",
                "min": 0,
                "format": "{:,.2f}",
                "scale": "Reds",
            },
            "max_umi_per_pos": {
                "title": "Max #UMI",
                "description": "Max number of unique UMIs per position",
                "min": 0,
                "format": "{:,.0f}",
                "scale": "Oranges",
            },
        }

        # add a section with a beeswarm plot of UMI stats to the report
        self.add_section(
            name="UMI Stats",
            anchor="umitools-umi-stats",
            description="Statistics from running `umi_tools dedup` or `umi_tools extract`",
            helptext="""
            - **Positions Dedup**: Total number of positions deduplicated
            - **Total UMIs**: Total UMIs found in sample
            - **Unique UMIs**: Unique UMIs found in sample
            - **Mean #UMI**: Mean number of unique UMIs per position
            - **Max #UMI**: Max number of unique UMIs per position
            """,
            plot=beeswarm.plot(
                data_by_sample,
                headers,
                {
                    "id": "umitools_stats_violin",
                    "table_title": "UMI-tools: UMI stats",
                },
            ),
        )



""" MultiQC module to parse output from HOMER tagdirectory """

import logging
import math
import os
import re

from multiqc.plots import bargraph, linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class TagDirReportMixin:
    def homer_tagdirectory(self):
        """Find HOMER tagdirectory logs and parse their data"""
        self.parse_gc_content()
        self.parse_re_dist()
        self.parse_tag_length_dist()
        self.parse_tagInfo_data()
        self.parse_freq_distribution_data()
        self.homer_stats_table_tagInfo()

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write data to file
        self.write_data_file(self.tagdir_data, "homer_tagdir")

        return sum([len(v) for k, v in self.tagdir_data.items() if k != "header"])

    def parse_gc_content(self):
        """parses and plots GC content and genome GC content files"""
        # Find and parse GC content:
        for f in self.find_log_files("homer/GCcontent", filehandles=True):
            # Get the s_name from the parent directory
            s_name = os.path.basename(f["root"])
            s_name = self.clean_s_name(s_name, f)
            parsed_data = self.parse_twoCol_file(f)
            if parsed_data is not None:
                if s_name in self.tagdir_data["GCcontent"]:
                    log.debug(f"Duplicate GCcontent sample log found! Overwriting: {s_name}")

                self.add_data_source(f, s_name, section="GCcontent")
                self.tagdir_data["GCcontent"][s_name] = parsed_data

        ## get esimated genome content distribution:
        for f in self.find_log_files("homer/genomeGCcontent", filehandles=True):
            parsed_data = self.parse_twoCol_file(f)
            if parsed_data is not None:
                if s_name + "_genome" in self.tagdir_data["GCcontent"]:
                    log.debug(f"Duplicate genome GCcontent sample log found! Overwriting: {s_name + '_genome'}")

                self.add_data_source(f, s_name + "_genome", section="GCcontent")
                self.tagdir_data["GCcontent"][s_name + "_genome"] = parsed_data

        self.tagdir_data["GCcontent"] = self.ignore_samples(self.tagdir_data["GCcontent"])

        if len(self.tagdir_data["GCcontent"]) > 0:
            self.add_section(
                name="Per Sequence GC Content",
                anchor="homer-tag-directory-gc-content-section",
                description="This plot shows the distribution of GC content.",
                plot=self.gc_content_plot(),
            )

    def parse_re_dist(self):
        """parses and plots restriction distribution files"""
        # Find and parse homer restriction distribution reports
        for f in self.find_log_files("homer/RestrictionDistribution", filehandles=True):
            s_name = os.path.basename(f["root"])
            s_name = self.clean_s_name(s_name, f)
            parsed_data = self.parse_restriction_dist(f)
            if parsed_data is not None:
                if s_name in self.tagdir_data["restriction"]:
                    log.debug(f"Duplicate Restriction Distribution sample log found! Overwriting: {s_name}")

                self.add_data_source(f, s_name, section="restriction")
                self.tagdir_data["restriction"][s_name] = parsed_data
                self.tagdir_data["restriction_norm"][s_name] = self.normalize(parsed_data)

        self.tagdir_data["restriction"] = self.ignore_samples(self.tagdir_data["restriction"])
        self.tagdir_data["restriction_norm"] = self.ignore_samples(self.tagdir_data["restriction_norm"])

        if len(self.tagdir_data["restriction"]) > 0:
            self.add_section(
                name="Restriction Site Tag Dist",
                anchor="homer-petag-restriction-dist-section",
                description="This plot shows the distribution of tags around restriction enzyme cut sites.",
                helptext="""Hi-C data often involves the digestion of DNA using a restriction enzyme.
                    A good quality control for the experiment is the centering of reads around the
                    restriction enzyme cut site.""",
                plot=self.restriction_dist_chart(),
            )

    def parse_tag_length_dist(self):
        """parses and plots tag length distribution files"""
        # Find and parse homer tag length distribution reports
        for f in self.find_log_files("homer/LengthDistribution", filehandles=True):
            s_name = os.path.basename(f["root"])
            s_name = self.clean_s_name(s_name, f)
            parsed_data = self.parse_length_dist(f)
            if parsed_data is not None:
                if s_name in self.tagdir_data["length"]:
                    log.debug(f"Duplicate Length Distribution sample log found! Overwriting: {s_name}")
                self.add_data_source(f, s_name, section="length")
                self.tagdir_data["length"][s_name] = parsed_data

        self.tagdir_data["length"] = self.ignore_samples(self.tagdir_data["length"])
        if len(self.tagdir_data["length"]) > 0:
            self.add_section(
                name="Tag Length Distribution",
                anchor="homer-tag-length-dist-section",
                description="This plot shows the distribution of tag length.",
                helptext="This is a good quality control for tag length inputed into Homer.",
                plot=self.length_dist_chart(),
            )

    def parse_tagInfo_data(self):
        """parses and plots taginfo files"""
        # Find and parse homer taginfo reports
        for f in self.find_log_files("homer/tagInfo", filehandles=True):
            s_name = os.path.basename(f["root"])
            s_name = self.clean_s_name(s_name, f)
            parsed_data = self.parse_tag_info_chrs(f)
            if parsed_data is not None:
                if s_name in self.tagdir_data["taginfo_total"]:
                    log.debug(f"Duplicate tag info sample log found! Overwriting: {s_name}")
                self.add_data_source(f, s_name, section="taginfo")
                self.tagdir_data["taginfo_total"][s_name] = parsed_data[0]
                self.tagdir_data["taginfo_total_norm"][s_name] = self.normalize(parsed_data[0])
                self.tagdir_data["taginfo_uniq"][s_name] = parsed_data[1]
                self.tagdir_data["taginfo_uniq_norm"][s_name] = self.normalize(parsed_data[1])

        for f in self.find_log_files("homer/tagInfo", filehandles=True):
            s_name = os.path.basename(f["root"])
            s_name = self.clean_s_name(s_name, f)
            ## collected tag_info data for general stats table and store under 'header'
            parsed_data = self.parse_tag_info(f)
            if parsed_data is not None:
                self.tagdir_data["header"][s_name] = parsed_data

        self.tagdir_data["taginfo_total"] = self.ignore_samples(self.tagdir_data["taginfo_total"])
        self.tagdir_data["taginfo_total_norm"] = self.ignore_samples(self.tagdir_data["taginfo_total_norm"])
        self.tagdir_data["taginfo_uniq"] = self.ignore_samples(self.tagdir_data["taginfo_uniq"])
        self.tagdir_data["taginfo_uniq_norm"] = self.ignore_samples(self.tagdir_data["taginfo_uniq_norm"])
        self.tagdir_data["header"] = self.ignore_samples(self.tagdir_data["header"])

        if len(self.tagdir_data["taginfo_total"]) > 0:
            self.add_section(
                name="Chromosomal Coverage",
                anchor="homer-tag-info-dist-section",
                description="This plot shows the distribution of tags along chromosomes.",
                helptext="""This is a good quality control for tag distribution and
                        could be a good indication of large duplications or deletions.""",
                plot=self.tag_info_chart(),
            )

    def parse_freq_distribution_data(self):
        """parses and plots taginfo files"""
        # Find and parse homer tag FreqDistribution_1000 reports
        for f in self.find_log_files("homer/FreqDistribution", filehandles=True):
            s_name = os.path.basename(f["root"])
            s_name = self.clean_s_name(s_name, f)
            parsed_data = self.parse_FreqDist(f)
            if parsed_data is not None:
                if s_name in self.tagdir_data["FreqDistribution"]:
                    log.debug(f"Duplicate Freq Distribution sample log found! Overwriting: {s_name}")

                self.add_data_source(f, s_name, section="FreqDistribution")
                self.tagdir_data["FreqDistribution"][s_name] = parsed_data

        self.tagdir_data["FreqDistribution"] = self.ignore_samples(self.tagdir_data["FreqDistribution"])

        if len(self.tagdir_data["FreqDistribution"]) > 0:
            self.add_section(
                name="Frequency Distribution",
                anchor="homer-freq-distribution-section",
                description="This plot shows the distribution of distance between PE tags.",
                helptext="""It is expected the the frequency of PE tags decays with
                    increasing distance between the PE tags. This plot gives an idea
                     of the proportion of short-range versus long-range interactions.""",
                plot=self.freq_dist_chart(),
            )

    def homer_stats_table_tagInfo(self):
        """Add core HOMER stats to the general stats table from tagInfo file"""

        if len(self.tagdir_data["header"]) == 0:
            return None

        headers = {
            "UniqPositions": {
                "title": "Uniq Pos",
                "description": "Numer of Unique Di-Tags Passed Through HOMER",
                "format": "{:,.0f}",
                "modify": lambda x: x * 0.000001,
                "suffix": "M",
            },
            "TotalPositions": {
                "title": "Total Pos",
                "description": "Numer of Total Di-Tags Passed Through HOMER",
                "format": "{:,.0f}",
                "modify": lambda x: x * 0.000001,
                "suffix": "M",
            },
            "fragmentLengthEstimate": {
                "title": "fragment Length",
                "description": "Estimate of Fragnment Length",
                "format": "{:,.0f}",
            },
            "peakSizeEstimate": {
                "title": "Peak Size",
                "description": "Estimate of Peak Size",
                "format": "{:,.0f}",
            },
            "tagsPerBP": {
                "title": "tagsPerBP",
                "description": "average tags Per basepair",
                "format": "{:,.3f}",
            },
            "TagsPerPosition": {
                "title": "averageTagsPerPosition",
                "description": "Average Tags Per Position",
                "format": "{:,.2f}",
            },
            "averageTagLength": {"title": "TagLength", "description": "Average Tag Length", "format": "{:,.0f}"},
            "averageFragmentGCcontent": {
                "title": "GCcontent",
                "description": "Average Fragment GC content",
                "max": 1,
                "min": 0,
                "format": "{:,.2f}",
            },
        }
        self.general_stats_addcols(self.tagdir_data["header"], headers)

    def homer_stats_table_interchr(self):
        """Add core HOMER stats to the general stats table from FrequencyDistribution file"""

        headers = {
            "InterChr": {
                "title": "InterChr",
                "description": "Fraction of Reads forming inter chromosomal interactions",
                "format": "{:,.4f}",
            }
        }
        self.general_stats_addcols(self.tagdir_data["FreqDistribution"], headers, "InterChr")

    def normalize(self, mydict, target=100):
        raw = sum(mydict.values())
        factor = target / raw
        return {key: value * factor for key, value in mydict.items()}

    def parse_twoCol_file(self, f):
        """Parse HOMER tagdirectory GCcontent file."""
        parsed_data = dict()
        firstline = True
        for line in f["f"]:
            if firstline:  # skip first line
                firstline = False
                continue
            s = line.split("\t")
            if len(s) > 1:
                k = float(s[0].strip())
                v = float(s[2].strip())
                parsed_data[k] = v

        return parsed_data

    def parse_restriction_dist(self, f):
        """Parse HOMER tagdirectory petagRestrictionDistribution file."""
        parsed_data = dict()
        firstline = True
        for line in f["f"]:
            if firstline:  # skip first line
                firstline = False
                continue
            s = line.split("\t")
            if len(s) > 1:
                nuc = float(s[0].strip())
                v1 = float(s[1].strip())
                v2 = float(s[2].strip())
                v = v1 + v2
                # parsed_data.update({nuc:v1})
                # parsed_data.update({nuc:v2})
                parsed_data.update({nuc: v})
        return parsed_data

    def parse_length_dist(self, f):
        """Parse HOMER tagdirectory tagLengthDistribution file."""
        parsed_data = dict()
        firstline = True
        for line in f["f"]:
            if firstline:  # skip first line
                firstline = False
                continue
            s = line.split("\t")
            if len(s) > 1:
                k = float(s[0].strip())
                v = float(s[1].strip())
                parsed_data[k] = v

        return parsed_data

    def parse_tag_info(self, f):
        """Parse HOMER tagdirectory taginfo.txt file to extract statistics in the first 11 lines."""
        # General Stats Table
        tag_info = dict()
        for line in f["f"]:
            s = line.split("=")
            if len(s) > 1:
                if s[0].strip() == "genome":
                    ss = s[1].split("\t")
                    if len(ss) > 2:
                        tag_info["genome"] = ss[0].strip()
                        try:
                            tag_info["UniqPositions"] = float(ss[1].strip())
                            tag_info["TotalPositions"] = float(ss[2].strip())
                        except Exception:
                            tag_info["UniqPositions"] = ss[1].strip()
                            tag_info["TotalPositions"] = ss[2].strip()
                try:
                    tag_info[s[0].strip()] = float(s[1].strip())
                except ValueError:
                    tag_info[s[0].strip()] = s[1].strip()
        return tag_info

    def parse_tag_info_chrs(self, f, convChr=True):
        """Parse HOMER tagdirectory taginfo.txt file to extract chromosome coverage."""
        parsed_data_total = {}
        parsed_data_uniq = {}
        remove = ["hap", "random", "chrUn", "cmd", "EBV", "GL", "NT_"]
        for line in f["f"]:
            s = line.split("\t")
            key = s[0].strip()
            # skip header
            if "=" in line or len(s) != 3:
                continue
            if convChr:
                if any(x in key for x in remove):
                    continue
            try:
                vT = float(s[1].strip())
                vU = float(s[2].strip())
            except ValueError:
                continue

            parsed_data_total[key] = vT
            parsed_data_uniq[key] = vU

        return [parsed_data_total, parsed_data_uniq]

    def parse_FreqDist(self, f):
        """Parse HOMER tagdirectory petag.FreqDistribution_1000 file."""
        parsed_data = dict()
        firstline = True
        for line in f["f"]:
            if firstline:
                firstline = False
                continue
            else:
                s = line.split("\t")
                if len(s) > 1:
                    k = s[0].strip()
                    if k.startswith("More than "):
                        k = re.sub("More than ", "", k)

                    k = float(k)
                    v = float(s[1].strip())
                    parsed_data[k] = v
        return parsed_data

    def parse_FreqDist_interChr(self, f):
        """Parse HOMER tagdirectory petag.FreqDistribution_1000 file to get inter-chromosomal interactions."""
        parsed_data = dict()
        firstline = True
        for line in f["f"]:
            if firstline:
                firstline = False
                interChr = float(re.sub(r"\)", "", line.split(":")[1]))
            else:
                break
        parsed_data["interChr"] = interChr
        return parsed_data

    def restriction_dist_chart(self):
        """Make the petagRestrictionDistribution plot"""

        pconfig = {
            "id": "homer-petag-restriction-dist",
            "title": "HOMER: Restriction Distribution",
            "ylab": "Reads",
            "xlab": "Distance from cut site (bp)",
            "data_labels": [{"name": "Number of Tags"}, {"name": "Percenatge"}],
        }
        datasets = [self.tagdir_data["restriction"], self.tagdir_data["restriction_norm"]]

        return linegraph.plot(datasets, pconfig)

    def length_dist_chart(self):
        """Make the tagLengthDistribution plot"""

        pconfig = {
            "id": "homer-tag-length-dist",
            "cpswitch": True,
            "title": "HOMER: Tag Length Distribution",
            "ylab": "Fraction of Tags",
            "xlab": "Tag Length (bp)",
        }
        return linegraph.plot(self.tagdir_data["length"], pconfig)

    def gc_content_plot(self):
        """Create the HTML for the Homer GC content plot"""

        pconfig = {
            "id": "homer-tag-directory-gc-content",
            "title": "HOMER: Tag Directory Per Sequence GC Content",
            "smooth_points": 200,
            "smooth_points_sumcounts": False,
            "ylab": "Normalized Count",
            "xlab": "% GC",
            "ymin": 0,
            "xmax": 1,
            "xmin": 0,
            "yDecimals": True,
            "tt_label": "<b>{point.x}% GC</b>: {point.y}",
        }
        return linegraph.plot(self.tagdir_data["GCcontent"], pconfig)

    def tag_info_chart(self):
        """Make the taginfo.txt plot"""

        ## TODO: human chrs on hg19. How will this work with GRCh genome or other, non human, genomes?
        # nice if they are ordered by size
        ucsc = ["chr" + str(i) for i in range(1, 23)].append(["chrX", "chrY", "chrM"])
        ensembl = list(range(1, 23)).append(["X", "Y", "MT"])
        pconfig = {
            "id": "homer-tag-info-dist",
            "title": "HOMER: Tag Info Distribution",
            "ylab": "Tags",
            "cpswitch_counts_label": "Number of Tags",
        }

        ## check if chromosomes starts with "chr" (UCSC) or "#" (ensembl)
        sample1 = next(iter(self.tagdir_data["taginfo_total"]))
        chrFormat = next(iter(self.tagdir_data["taginfo_total"][sample1]))

        if "chr" in chrFormat:
            chrs = ucsc
        else:
            chrs = ensembl

        return bargraph.plot(self.tagdir_data["taginfo_total"], chrs, pconfig)

    def freq_dist_chart(self):
        """Make the petag.FreqDistribution_1000 plot"""
        # Take a log of the data before plotting so that we can
        # reduce the number of points to plot evenly
        pdata = {}
        for idx, s_name in enumerate(self.tagdir_data["FreqDistribution"]):
            pdata[s_name] = {}
            for x, y in self.tagdir_data["FreqDistribution"][s_name].items():
                try:
                    pdata[s_name][math.log(float(x))] = y
                except ValueError:
                    pass
        pconfig = {
            "id": "homer-freq-distribution",
            "title": "HOMER: Frequency Distribution",
            "ylab": "Fraction of Reads",
            "xlab": "Log10(Distance between regions)",
            "data_labels": ["Reads", "Percent"],
            "smooth_points": 500,
            "smooth_points_sumcounts": False,
            "yLog": True,
        }
        return linegraph.plot(pdata, pconfig)



import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Filtlong",
            anchor="filtlong",
            href="https://github.com/rrwick/Filtlong",
            info="A tool for filtering long reads by quality.",
            # doi="", # No DOI
        )

        # Find and load reports
        self.filtlong_data = dict()

        # Find all files for filtlong
        for f in self.find_log_files("filtlong", filehandles=True):
            self.parse_logs(f)

        self.filtlong_data = self.ignore_samples(self.filtlong_data)

        if len(self.filtlong_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.filtlong_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write data to file
        self.write_data_file(self.filtlong_data, "filtlong")
        self.filtlong_general_stats()

    def parse_logs(self, f):
        """Parsing Logs. Note: careful of ANSI formatting log"""
        for line in f["f"]:
            # Find the valid metric
            if "target:" in line:
                self.add_data_source(f)
                if f["s_name"] in self.filtlong_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                target_bases = line.lstrip().split(" ")[1]
                # Remove . thousand separators - see MultiQC/MultiQC#1843
                target_bases = float(target_bases.replace(".", ""))
                self.filtlong_data[f["s_name"]] = {"Target bases": target_bases}

            elif "keeping" in line and f["s_name"] in self.filtlong_data:
                bases_kept = line.lstrip().split(" ")[1]
                # Remove . thousand separators - see MultiQC/MultiQC#1843
                bases_kept = float(bases_kept.replace(".", ""))
                self.filtlong_data[f["s_name"]]["Bases kept"] = bases_kept

            elif "fall below" in line and f["s_name"] in self.filtlong_data:
                log.debug(f"{f['s_name']}: reads already fall below target after filtering")

    def filtlong_general_stats(self):
        """Filtlong General Stats Table"""
        headers = {
            "Target bases": {
                "title": f"Target bases ({config.read_count_prefix})",
                "description": "Keep only the best reads up to this many total bases ({})".format(
                    config.read_count_desc
                ),
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "Bases kept": {
                "title": f"Bases kept ({config.read_count_prefix})",
                "description": f"Bases kept ({config.read_count_desc})",
                "scale": "Purples",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
        }

        self.general_stats_addcols(self.filtlong_data, headers)
